"""
SHEPHARD: 
Sequence-based Hierarchical and Extendable Platform for High-throughput Analysis of Region of Disorder

Authors: Garrett M. Ginell & Alex S. Holehouse
Contact: (alex.holehouse@wustl.edu, g.ginell@wustl.edu)

Holehouse Lab - Washington University in St. Louis
"""


from shephard.exceptions import InterfaceException, ProteinException, TrackException
import shephard.exceptions as shephard_exceptions
from . import interface_tools 
from shephard import general_utilities
import os



class _TracksInterface:

    def __init__(self, filename, delimiter='\t', mode='values', skip_bad=True):
        """
        
        Class for reading in correctly formatted tracks files for parsing 
        into a Proteome object.

        Tracks files must adhere to the following specification

            unique_ID, track_name, val_1, val_2, ...., val_n 

        where n = length of protein.

        This class allows a tracksfile to be read in and defined as either 
        a values track file, or a symbols track file, returning a tracks 
        dictionary. 

        Parameters
        ----------------
        
        filename : str
            Name of the SHEPHARD Tracks file to read.


        Other Parameters
        ----------------

        delimiter : str (default = '\\t')
            String used as a delimiter on the input file. 

        mode : str (default = 'values')
            A selector that defines the type of track file to be read. 
            Must be either 'symbols' or 'values'.

        skip_bad : bool (default = True)
            Flag that means if bad lines (lines that trigger an exception) 
            are encountered the code will just skip them. By default this is 
            true, which adds a certain robustness to file parsing, but could 
            also hide errors. Note that if lines are skipped a warning will 
            be printed (regardless of verbose flag). 

        """
        
        with open(filename,'r') as fh:
            content = fh.readlines()

        ID2track = {}

        linecount = 0

        # cycle over every line in the file
        for line in content:

            linecount = linecount + 1

            # skip comment lines
            if interface_tools.is_comment_line(line):
                continue

            # extract chop off lagging whitespace and divide up using the delimiter
            sline = line.strip().split(delimiter)                        
            track_data = []
            
            # for this list 
            try:

                # extract track name and unique_id
                unique_ID = sline[0].strip()
                track_name = sline[1].strip()
                
                # parse track values or symbols
                if mode == 'values':

                    # for each element in sline strip whitespace and convert to a float
                    track_data = [float(i.strip()) for i in sline[2:]]

                elif mode == 'symbols':
                    # for each element in sline strip whitespace 
                    track_data = [i.strip() for i in sline[2:]]
                else:
                    raise InterfaceException('Error: %s'% "mode passed = %s, yet this does not match 'symbols' or 'values'")

                        
                if unique_ID in ID2track:
                    ID2track[unique_ID].append({'track_name':track_name, 'track_data':track_data})                    
                else:
                    ID2track[unique_ID] = [{'track_name':track_name, 'track_data':track_data}]
                    

            except Exception as e:

                msg = 'Failed parsing file [%s] on line [%i].\n\nException raised: %s\n\nline printed below:\n%s'%(filename, linecount, str(e), line)

                # should update this to also display the actual error...
                if skip_bad:
                    shephard_exceptions.print_warning(msg + "\nSkipping this line...")
                    continue
                else:
                    raise InterfaceException(msg)

        self.data = ID2track


## ------------------------------------------------------------------------
##
def __write_all_tracks_single_file(proteome, 
                                   outfile, 
                                   track_type,
                                   value_fmt = "%.3f", 
                                   delimiter='\t'):
    r"""
    Internal function Function that writes all tracks associated with  
    a Proteome out to a single file. 

    See also:

        write_all_values_tracks_single_file() 
        write_all_symbols_tracks_single_file() 
    

    Parameters
    -----------

    proteome : Proteome object
        Proteome object from which the Domains will be extracted from

    outfile : str
        String that defines the name of the output file.


    Other Parameters
    ----------------

    value_fmt : str (default = "%.3f")
        Format string that will be used for values. 
        
    delimiter : str (default = '\\t')
        Character (or characters) used to separate between fields. 
        Default is '\t' which is recommended to maintain compliance 
        with default `add_tracks_from_files()` function.

    
    Returns
    --------
    None
        No return type, but generates a new file with the complete 
        set of tracks from this Proteome written to disk.
        

    """
    
    # open the file handle
    fh = open(outfile,'w')

    # build a list of track names that are values-tracks
    tn2tt = proteome.track_names_to_track_type
    
    valid_names = []
    for name in tn2tt:
        if tn2tt[name] == track_type:
            valid_names.append(name)
        
    # cyle through each track name we designated as valid
    for t_name in valid_names:        
        write_tracks(proteome, None, t_name, value_fmt, delimiter, file_handle=fh)

    fh.close()


##############################################
##                                          ##
##     PUBLIC FACING FUNCTIONS BELOW        ##
##                                          ##
##############################################

## ------------------------------------------------------------------------
##
def add_tracks_from_file(proteome, filename, mode, delimiter='\t', return_dictionary=False, safe=True, skip_bad=True, verbose=True):
    r"""
    Function that takes a correctly formatted shephard 'tracks' file and reads 
    all Tracks into the passed Proteome.

    Expect Track files to have the following format:

    One protein per line, where each protein has the following information:
    
    >>> Unique_ID    track_name    res1    res2    res3 .... resn

    Where ``res1``, ``res2``, ``resn`` are symbol or values to be mapped to 
    the 1st, 2nd, or nth residue. There should be the same number of res1, 2, 
    ...n entries are there are residues in the associated protein.
    
    A couple of key points here:

    - The default delimiter is tabs ('\\t') but this can be changed with 
      the delimiter argument

    - Each track must assign a value or a symbol to EVERY residue in the 
      protein
    
    Parameters
    ----------

    proteome : Proteome
        Proteome object 

    filename : str
        Name of the shephard Domains file to read

    mode : str (default = 'values')
        A selector that defines the type of track file to be read. 
        Must be either 'symbols' or 'values'.

    delimiter : str (default = '\\t')
        String used as a delimiter on the input file. 

    return_dictionary : bool (default = False)
        If set to true, this function will return the tracks dictionary 
        and will NOT add that dictionary to the Proteome - i.e. the function 
        basically becomes a parser for SHEPHARD-compliant tracks files. 
        
    safe : bool (default = True)
        If set to True then any exceptions raised during the Track-adding 
        process (i.e. after file parsing) are acted on. If set to False, 
        exceptions simply mean the site in question is skipped. Note if set 
        to False pre-existing tracks with the same name would be silently 
        overwritten (although this is not consider an error), while 
        overwriting will trigger an exception in safe=True. There are various 
        reasons site addition could fail (e.g. track does not match length of 
        protein) so if verbose=True then the cause of an exception is also 
        printed to screen. It is highly recommend that if you choose to use 
        safe=False you also set verbose=True. 
        
    skip_bad : bool (default = True)
        Flag that means if bad lines (lines that trigger an exception) are 
        encountered the code will just skip them. By default this is true, 
        which adds a certain robustness to file parsing, but could also hide 
        errors. Note that if lines are skipped a warning will be printed 
        (regardless of verbose flag). 

    verbose : bool (default = True)
        Flag that defines how 'loud' output is. Will warn about errors on 
        adding tracks.

    Returns
    -----------
    None or dict
        If return_dictionary is set to False (default) then this function 
        has no return value, but the tracks are added to the Proteome object
        passed as the first argument. If return_dictionary is set to True the 
        function returns the parsed tracks dictionary without adding the 
        newly-read tracks to the proteome.
        
    """        

    # check first argument is a Proteome
    interface_tools.check_proteome(proteome, 'add_tracks_from_file (si_tracks)')

    # check mode is valid
    general_utilities.valid_keyword('mode', mode, ['symbols','values'])
    
    # next read in the file
    tracks_interface = _TracksInterface(filename, delimiter=delimiter, mode=mode, skip_bad=skip_bad)

    if return_dictionary:
        return tracks_interface.data

    # finally add the domains from the dictionary generated by the DomainsInterface parser
    add_tracks_from_dictionary(proteome, tracks_interface.data, mode, safe=safe, verbose=verbose)



## ------------------------------------------------------------------------
##
def add_tracks_from_dictionary(proteome, tracks_dictionary, mode, safe=True, verbose=True):
    """
    Function that takes a correctly formatted Tracks dictionary and 
    will add those Tracks to the proteins in the Proteome.
    
    Track dictionaries are key-value pairs, where the key is a unique 
    ID and the value is a list of dictionaries. For each sub-dictionary, 
    there are two key-value pairs that reflect:

       * 'track_name'  : name of the track (str)

       * 'track_data'  : parsed list of floats (if expecting values) or strings (if expecting symbols) that should equal the length of the associated protein.
                       
    Parameters
    ------------

    proteome : Proteome Object
        Proteome object which tracks will be added to

    tracks_dictionary : dict
        Dictionary in which keys are unique IDs for proteins and the value 
        is a list of dictionaries, where each subdictionary has the two 
        key-value pairs:
        
        * **track_name**  : name of the track (str)
        * **track_data**  : parsed list of floats (if expecting values) or strings (if expecting symbols) that should equal the length of the associated protein.
                            
    mode : str (default = 'values')
        A selector that defines the type of track file to be read. 
        Must be either 'symbols' or 'values'.

    safe : bool (default = True)
        If set to True then any exceptions raised during the track-adding 
        process are acted on. If set to False, exceptions simply mean the 
        Track in question is skipped. Note if set to False, pre-existing 
        Tracks with the same name would be silently overwritten (although 
        this is not consider an error), while overwriting will trigger an 
        exception in safe=True. There are various reasons Track addition 
        could fail (length does not match the protein etc) and so if 
        verbose=True then the cause of an exception is also printed to 
        screen. It is highly recommend that if you choose to use 
        safe=False you also set verbose=True. 

    verbose : boolean (default = True)
        Flag that defines how 'loud' output is. Will warn about errors on 
        adding tracks.
        
    Returns
    -----------
    None
        No return value, but tracks are added to the Proteome object 
        passed as the first argument.
        
    """
        
    # check first argument is a proteome
    interface_tools.check_proteome(proteome, 'add_tracks_from_dictionary (si_tracks)')

    # check mode is valid
    general_utilities.valid_keyword('mode', mode, ['symbols','values'])
    
    # cycle through each protein in the proteome...
    for protein in proteome:
        if protein.unique_ID in tracks_dictionary:
            for track in tracks_dictionary[protein.unique_ID]:

                # get the track name and vector info
                track_name = track['track_name']
                track_data = track['track_data']

                # add the track as either values or symbols depending 
                # on what was provided
                try:
                    if mode == 'values':
                        protein.add_track(track_name, values=track_data, safe=safe)
                    else:
                        protein.add_track(track_name, symbols=track_data, safe=safe)

                # if an ProteinException was raised when trying to add a track some
                # anticipated error occurred
                except (ProteinException, TrackException) as e:      

                    msg='- skipping track at %s on %s' %(track_name, protein)
                    if safe:
                        shephard_exceptions.print_and_raise_error(msg, e)
                    else:
                        if verbose:
                            shephard_exceptions.print_warning(msg)
                        continue





## ------------------------------------------------------------------------
##
def write_all_tracks_separate_files(proteome, 
                     outdirectory='.', 
                     value_fmt = "%.3f", 
                     delimiter='\t'):

    """
    Function that writes all tracks associated with a proteome out to seperate
    files. This may be preferable in some situations, but in others maybe only
    a subset of tracks are requested, for which write_tracks() would be good, 
    or alternatively you want all tracks in a single file, in which case
    write_all_tracks_single_files() would be the way to go.

    The the output filenames are defined as:
        
    > `shephard_track_<trackname>.tsv`
    
    and are written to the outdirectory.

    Because track files MUST be written as one per track_name, this 
    function is equivalent to cycling through each unique track name 
    and writing it out sequentially. 
    
    Parameters
    -----------

    proteome :  Proteome object
        Proteome object from which the Domains will be extracted from

    outdirectory : str (default = '.')
        String that defines the output directory. By default sets to the 
        present working directory.

    value_fmt : str (default = "%.3f")
        Format string that will be used for values. Default = "%.3f"
        
    delimiter : str (default = '\\t')
        Character (or characters) used to separate between fields. 
        Default is '\t' Which is recommended to maintain compliance with 
        default `add_tracks_from_files()` function.
    
    Returns
    --------
    None
        No return type, but generates a new file with the complete set of 
        Domains from this Proteome written to disk.

    """

    for t_name in proteome.unique_track_names:
        outname = os.path.join(outdirectory, "shephard_track_%s.tsv" %( t_name))            
        write_tracks(proteome, outname, t_name, value_fmt, delimiter)


## ------------------------------------------------------------------------
##
def write_all_values_tracks_single_file(proteome, 
                                 outfile, 
                                 value_fmt = "%.3f", 
                                 delimiter='\t'):
    r"""
    Function that writes all tracks associated with a Proteome out to a single
    file. This may be preferable in some situations, but in others maybe only
    a subset of tracks are requested, for which write_tracks() would be good, 
    or alternatively you want all tracks in seperate files, in which case
    write_all_tracks_separate_files() would be the way to go.
        
    Parameters
    -----------

    proteome :  Proteome object
        Proteome object from which the Domains will be extracted from

    outfile : str
        String that defines the name of the output file.

    value_fmt : str (default = "%.3f")
        Format string that will be used for values. Default = "%.3f"
        
    delimiter : str (default = '\\t')
        Character (or characters) used to separate between fields. Default 
        is '\t' Which is recommended to maintain compliance with default 
        `add_tracks_from_files()` function.
    
    Returns
    --------
    None
        No return type, but generates a new file with the complete set of tracks
        from this Proteome written to disk.


    """

    return __write_all_tracks_single_file(proteome, outfile, 'values', value_fmt, delimiter)

## ------------------------------------------------------------------------
##
def write_all_symbols_tracks_single_file(proteome, 
                                 outfile, 
                                 value_fmt = "%.3f", 
                                 delimiter='\t'):
    r"""
    Function that writes all tracks associated with a Proteome out to a single
    file. This may be preferable in some situations, but in others maybe only
    a subset of tracks are requested, for which write_tracks() would be good, 
    or alternatively you want all tracks in seperate files, in which case
    write_all_tracks_separate_files() would be the way to go.
        
    Parameters
    -----------

    proteome :  Proteome object
        Proteome object from which the Domains will be extracted from

    outfile : str
        String that defines the name of the output file.

    value_fmt : str (default = "%.3f")
        Format string that will be used for values.
            
    delimiter : str (default = '\\t')
        Character (or characters) used to separate between fields. Default
        is '\t' Which is recommended to maintain compliance with default 
        `add_tracks_from_files()` function.
    
    Returns
    --------
    None
        No return type, but generates a new file with the complete set of 
        tracks from this proteome written to disk.


    """
    return __write_all_tracks_single_file(proteome, outfile, 'symbols', value_fmt, delimiter)



## ------------------------------------------------------------------------
##
def write_tracks(proteome, filename, track_name, value_fmt = "%.3f", delimiter='\t', file_handle=None):
    r"""
    Function that writes out a specific track to file in a standardized 
    format. Note that because track files are inevitably quite big default 
    behaviour is to only write out a single track file at a time (i.e. 
    unlike write_domains or write_sites where ALL domains or all sites are
    - by default - written out, here ONLY a single type of track, defined
    by track_name, can be written.

    To write ALL the tracks from a file, see si_tracks.write_all_tracks().
    
    Parameters
    -----------
    proteome :  Proteome object
        Proteome object from which the Domains will be extracted from

    filename : str
        Filename that will be used to write the new Domains file

    track_name : str
        Name of the track to be written out.

    value_fmt : str (default = "%.3f")
        Format string that will be used for values. Default = "%.3f". Note 
        that this is not a smart value so if the actual value used means 
        that %.3f looses all meaning this will not trigger a warning, so, 
        be careful!
        
    delimiter : str (default = '\\t')
        Character (or characters) used to separate between fields. Default 
        is '\t' which is recommended to maintain compliance with default 
        `add_tracks_from_files()` function.

    file_handle : filehandle (_io.TextIOWrapper) or None
        If passed, output is written to this handle rather than to a new 
        file. The filename variable is ignored in this case.
        
    
    Returns
    --------
    None
        No return type, but generates a new file with the complete set of 
        Domains from this Proteome written to disk.

    """

    # test the passed value_fmt string works. This is not fullproof but at least validates that
    # the string can parse a float (which is a necessary requirement for tracks values to be read
    # back in again by shephard
    try:
        a = value_fmt %( 1.5 )

        if float(a) != 1.5:
            raise InterfaceException('Invalid value_fmt passed [%s]'%(str(value_fmt)))
    except TypeError:
        raise InterfaceException('Invalid value_fmt passed [%s]'%(str(value_fmt)))

    if file_handle is not None:
        fh  = file_handle
    else:
        fh = open(filename, 'w')

    for protein in proteome:

            
        # try and extract out the track in question
        t = protein.track(track_name, safe=False)
        if t is not None:

            # construct an outstring from the track
            outstring = __build_track_line(t, delimiter, value_fmt)
            fh.write(outstring)

    if file_handle is None:
        fh.close()
                    
## ------------------------------------------------------------------------
##
def __build_track_line(t, delimiter, value_fmt):
    """
    Internal function that takes a Track object and returns a line that can
    be written to a Sites file. This is called internally by functions that
    write Tracks.

    Parameters
    ----------------------
    t : shephard.Track
        Track object being converted to a string

    delimiter : str (default = '\\t')
        Character (or characters) used to separate between fields. 
        Default is the tab character ('\\t'), which is recommended to 
        maintain compliance with default SHEPHARD file-reading functions.     

    Returns
    --------------
    str
        Returns a string that is ready to be written to file

    """

    unique_ID = t.protein.unique_ID

    # build the initial string
    #out_string = "%s%s%s%s" % (unique_ID, delimiter, t.name, delimiter)

    # build the initial string
    out_string = f"{unique_ID}{delimiter}{t.name}{delimiter}"

    if t.values is not None:
        for v in t.values:
            out_string = out_string + "%s%s" % (value_fmt %(v), delimiter)
    else:
        for v in t.symbols:
            out_string = out_string + "%s%s" % (v, delimiter)


    return out_string + "\n"
                    

## ------------------------------------------------------------------------
##
def write_tracks_from_list(track_list, filename, value_fmt = "%.3f", delimiter='\t'):
    r"""
    Function that writes out tracks to a SHEPHARD tracks file from a list
    of Track objects. 

    Note that attributes are converted to a string, which for simple 
    attributes is reasonable but is not really a viable stratergy for 
    complex objects, although this will not yeild and error.

    Note also that a single track file cannot have both values and
    symbols tracks, and this is checked first 
            
    Parameters
    -----------

    track_list : List of Track objects
        List of track objects which will be written

    filename : str
        Filename that will be used to write the new tracks file

    value_fmt : str (default = "%.3f")
        Format string that will be used for values. Default = "%.3f". Note 
        that this is not a smart value so if the actual value used means 
        that %.3f looses all meaning this will not trigger a warning, so, 
        be careful!

    delimiter : str (default = '\\t')
        Character (or characters) used to separate between fields. Default is 
        '\\t' which is recommended to maintain compliance with default 
        `add_tracks_from_file()` function
       
    Returns
    --------
    None
        No return type, but generates a new file with the complete set of 
        tracks from this proteome written to disk.

    """

    # first check if items in the list are track objects
    for t in track_list:
        interface_tools.check_track(t, 'write_tracks_from_list')

    ## START OF SANITY CHECKING FOR CONSISTENT TRACK TYPES
    ##
    ## The code below checks all tracks in the list are consistent, i.e.
    ## that they are ALL symbol or ALL value tracks
    ##

    # if none then this must be a symbol track
    if track_list[0].values is None:
        symbol = True
        values = False

    # else this must be a value track
    else:
        symbol = False
        values = True
        
    # make sure all tracks are consistent!
    for t in track_list:
        if symbol is True:
            if t.values is not None:
                raise InterfaceException(f'First track [{track_list[0]}] was a symbols track, yet track {t} is a values track. All tracks in the list must be the same type')

        if values is True:
            if t.symbols is not None:
                raise InterfaceException(f'First track [{track_list[0]}] was a values track, yet track {t} is a symbols track. All tracks in the list must be the same type')
            
    ## END OF SANITY CHECKING FOR CONSISTENT TRACK TYPES


    # test the passed value_fmt string works. This is not fullproof but at least validates that
    # the string can parse a float (which is a necessary requirement for tracks values to be read
    # back in again by shephard
    try:
        a = value_fmt %( 1.5 )

        if float(a) != 1.5:
            raise InterfaceException(f'Invalid value_fmt passed [{value_fmt}]')
    except TypeError:
        raise InterfaceException(f'Invalid value_fmt passed [{value_fmt}]')

            
    with open(filename, 'w') as fh:

        # for each track in the list
        for t in track_list:

            # build a line 
            # if the passed parameter track_types is being
            # used
            outstring = __build_track_line(t, delimiter, value_fmt)

            fh.write(outstring)

    





"""
SHEPHARD: 
Sequence-based Hierarchical and Extendable Platform for High-throughput Analysis of Region of Disorder

Authors: Garrett M. Ginell & Alex S. Holehouse
Contact: (g.ginell@wustl.edu)

Holehouse Lab - Washington University in St. Louis

"""

def print_warning(msg):
    """
    Function that prints a warning message (with an exception if provided).
    """
    
    print('WARNING: %s' %(msg))


def print_and_raise_error(msg, e):    
    print('ERROR: %s' %(msg))
    raise e

    

class ShephardException(Exception):
    """
    General exception
    """
    pass


# <><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><>
#
class SiteException(Exception):
    """
    Exception for the Site class
    """
    pass

# <><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><>
#
class TrackException(Exception):
    """
    Exception for the Domain class
    """
    pass


# <><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><>
#
class DomainException(Exception):
    """
    Exception for the Domain class
    """
    pass


# <><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><>
#
class ProteinException(Exception):
    """
    Exception for the Proteins class
    """
    pass


# <><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><>
#
class ProteomeException(Exception):
    """
    Exception for the Proteome class
    """
    pass


# <><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><>
#
class UtilitiesException(Exception):
    """
    Exception for general utility exceptions
    """
    pass


# <><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><>
#
class InterfaceException(Exception):
    """
    Exception for general utility exceptions
    """
    pass


# <><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><>
#
class APIException(Exception):
    """
    Exception for general utility exceptions
    """
    pass



"""
SHEPHARD: 
Sequence-based Hierarchical and Extendable Platform for High-throughput Analysis of Region of Disorder

Authors: Garrett M. Ginell & Alex S. Holehouse
Contact: (g.ginell@wustl.edu)

Holehouse Lab - Washington University in St. Louis
 

"""

from . import sequence_utilities
from .exceptions import DomainException
from . import general_utilities
from .tools import domain_tools

# <><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><>
# Class that defines a sequence region 
#
class Domain:

    ## ------------------------------------------------------------------------
    ##          
    def __init__(self, start, end, protein, domain_type, domain_name, attributes=None):
        """
        Domains are defined sub-regions within a protein. 
    
        Proteins contain a list of 0 or more domains, and each domain is 
        associated with the protein it originates from via the linking 
        protein object.
        
        Domains are indexed using the same indexing as the overall protein 
        sequence (i.e. a protein does not automatically start from 1), and 
        as such the 'native' sequence indexing should be used for working 
        with Proteins. This is a long-winded way of saying position X refers 
        to the same residue regardless of if it's taken from the Protein 
        or Domain or Track object.
        
        Parameters
        -------------

        start : int
            Start position in sequence (recall we index from 1)

        end : int
            End position in sequence (recall we index from 1)

        protein : Protein
            Protein object for which this Domain is part of

        domain_type :  str
            Name of the domain type - can be any free-form string

        domain_name : str
            The name used as an index by the associated protein to identify 
            this domain. This is ONLY used such that you can re-reference a 
            domain back to the protein if needed.

        attributes : dict (default = None)
            Dictionary where key/value pairs allow a Domain to have
            arbitrary metadata associated with it.        
        """
        
        if start > end:
            raise DomainException("Trying to a domain to protein [%s] where start site is bigger than the end site (positions: %i-%i - this does not work!" %(str(protein), start, end))

        # check the domain falls within the region
        helper_string="Trying to add domain to protein [%s] at positions [%i-%i] - this falls outside the protein's dimensions [%i-%i]" %(protein, start, end, 1, protein._len)
        protein._check_position_is_valid(start, helper_string)
        protein._check_position_is_valid(end, helper_string) 

        # assign if all OK
        self._start = int(start)
        self._end   = int(end)

        self._protein = protein         
        self._domain_type = domain_type
        self._domain_name = domain_name

        general_utilities.variable_is_dictionary(attributes, DomainException, 'attributes argument passed to domain %s [%i-%i] in protein %s is not a dictionary' %(self._domain_type, self._start, self._end, self._protein), or_none=True)

        if attributes is None:
            self._attributes = {}
        else:
            self._attributes = attributes

        # update unique domain types
        protein.proteome.__update_domain_types(self._domain_type)


        
    ## ------------------------------------------------------------------------
    ##
    @property
    def attributes(self):
        """
        Provides a list of the keys associated with every attribute associated
        with this domain.

        Returns
        -------
        list
            returns a list of the attribute keys associated with the domain. 


        """
        return list(self._attributes.keys())
    

    
    ## ------------------------------------------------------------------------
    ##
    def attribute(self, name, safe=True):
        """
        Function that returns a specific attribute as defined by the name. 

        Recall that attributes are name : value pairs, where the 'value' 
        can be anything and is user defined. This function will return 
        the value associated with a given name.

        Parameters
        ----------------
        name : str
             The attribute name. A list of valid names can be found by 
             calling the ``<domain>.attributes()`` (which returns a list 
             of the valid names)

        safe : bool (default = True)
            Flag which if true with throw an exception if an attribute 
            with the same name already exists.
            
        Returns
        ---------
        Unknown 
            Will either return whatever was associated with that attribute 
            (which could be anything) or None if that attribute is missing.
        """

        # if name is in the _atributes dictionary the  return
        if name in self._attributes:
            return self._attributes[name]
        else:

            # else if safe was passed raise an exception if that attribute was missing
            if safe:
                raise DomainException('Requesting attribute [%s] from domain [%s] but this attribute has not been assigned' % (name, str(self))) 

            # if safe not passed just return None
            else:
                return None                



    ## ------------------------------------------------------------------------
    ##
    def add_attribute(self, name, val, safe=True):
        """
        Function that adds an attribute. Note that if safe is true, this 
        function will raise an exception if the attribute is already present. 
        If safe=False, then an exisiting value will be overwritten.

        Parameters
        ----------------

        name : str
            Name that will be used to identify the attribute

        val : <anything>
            An object or primitive we wish to associate with this attribute

        safe : bool (default = True)
            Flag which if True with throw an exception if an attribute with 
            the same name already exists, otherwise the newly introduced 
            attribute will overwrite the previous one.

        Returns
        ---------
            None - but adds an attribute to the calling object

        """

        if safe:
            if name in self._attributes:
                raise DomainException("Trying to add attribute [%s=%s] to domain [%s] but this attribute is already set.\nPossible options are: %s" %(name,val, str(self), str(self._attributes.keys())))
                
        self._attributes[name] = val 



    ## ------------------------------------------------------------------------
    ##
    def remove_attribute(self, name, safe=True):
        """
        Function that removes a given attribute from the Domain based on the 
        passed attribute name. If the passed attribute does not exist or is not 
        associate with the Domain then this will trigger an exception 
        unless safe=False.

        Parameters
        ----------------

        name : str
            The attribute name that will be used to identify it

        safe : bool (default = True)
            Flag which if True with throw an exception if an 
            attribute this name does not exists. If set to
            False then if an attribute is not found it is simply
            ignored
            
        Returns
        ---------
        None
            No return type but will remove an attribute from the 
            protein if present.
            
        """

        if name not in self._attributes:
            if safe:
                raise DomainException(f'Passed attribute [{name}] not found in {self}')
        else:
            del self._attributes[name]


    ## ------------------------------------------------------------------------
    ##      
    @property
    def start(self):
        """
        **[Property]**: Returns the start position that defines this domain

        :getter: Returns the start of the domain (indexed from 1) 
        :setter: None
        :type: int

        """

        return self._start



    ## ------------------------------------------------------------------------
    ##      
    @property
    def end(self):
        """
        **[Property]**: Returns the end position that defines this domain
        """
        return self._end



    ## ------------------------------------------------------------------------
    ##      
    @property
    def protein(self):
        """
        **[Property]**: Returns the Protein that this Domain is associated 
        with
        """
        return self._protein



    ## ------------------------------------------------------------------------
    ##      
    @property
    def sequence(self):
        """
        **[Property]**: Returns the amino acid sequence associated with 
        this domain
        """
        return self._protein.get_sequence_region(self._start, self._end)



    ## ------------------------------------------------------------------------
    ##      
    @property
    def domain_type(self):
        """
        Returns the domain type as a string
        """
        return self._domain_type



    ## ------------------------------------------------------------------------
    ##      
    @property
    def domain_name(self):
        """
        Returns the domain name as generated when added to the protein
        """
        return self._domain_name



    ## ------------------------------------------------------------------------
    ##      
    def update_domain_name(self, new_name):
        """
        Function that updates the domains name
        
        Parameters
        -----------
        new_name : str
            String passed as new name

        Returns
        -----------
        None
            Nothing but sets the name to be the new name

        """
        self._domain_name = new_name




    ######################################
    ##                                  ##
    ##     DOMAIN  FUNCTIONS             #
    ##                                  ##
    ######################################

    
    ## ------------------------------------------------------------------------
    ##      
    def inside_domain(self, position):
        """
        Function that returns True/False depending on if the provided position
        lies inside the domain.

        Parameters
        ------------
        position : int
            Position in the sequence

        Returns
        -----------
        bool
            Returns True if position is inside the domain region, else False
        

        """
        return sequence_utilities.inside_region(self.start, self.end, position)


    ## ------------------------------------------------------------------------
    ##      
    def domain_overlap(self, domain2):
        """
        Function that takes in a second domain and calculates if those two 
        domains overlap at all. This is a binary check and does not compute 
        the extent of overlap.

        Parameters
        ------------
        domain2 : Domain
            The Domain object of interest

        Returns
        -----------
        bool
            Returns True if the domains overlap, False if not. Note this 
            will throw an exception if the domains are from different 
            proteins.

        """
        return domain_tools.domain_overlap(self, domain2)


    ######################################
    ##                                  ##
    ##     DOMAIN SITE FUNCTIONS        ##
    ##                                  ##
    ######################################

    ## ------------------------------------------------------------------------
    ##
    @property
    def sites(self):
        """
        Get list of all sites inside the domain.
        
        Returns
        --------
        list
            Returns a list of all the sites 
        """
        
        all_sites = []
        sites_dict = self._protein.get_sites_by_range(self.start, self.end)
        for k in sites_dict:
            for local_site in sites_dict[k]:
                all_sites.append(local_site)

        return all_sites


    ## ------------------------------------------------------------------------
    ##
    @property
    def site_positions(self):
        """
        Get list of all sites inside the domain.
        
        Returns
        --------
        list
            Returns a list of all the site positions
        """
        return list(self._protein.get_sites_by_range(self.start, self.end).keys())


    ## ------------------------------------------------------------------------
    ##
    def site(self, position):
        """
        Returns the list of sites that are found at a given position. Note 
        that - in generalsite() should be used to retrieve sites you know 
        exist while get_sites_by_position() offers a way to more safely get 
        sites at a position. Site will throw an exception if the position 
        passed does not exist (while get_sites_by_position() will not).

        Parameters
        -------------
        position : int
            Defines the position in the sequence we want to interrogate 

        Returns
        ---------
        list
            Returns a list with between 1 and n sites. Will raise an 
            exception if the passed position cannot be found in the 
            codebase.
        """

        ipos = int(position)

        if sequence_utilities.inside_region(self.start, self.end, ipos):
            return self._protein._sites[int(position)]
        else:
            raise DomainException('Passed position [%i] is outside of the domain boundaries [%i-%i]' %(ipos, domain.start, domain.end))


    ## ------------------------------------------------------------------------
    ##
    def get_sites_by_type(self, site_type, return_list=False):
        """
        Get dictionary of list of sites inside the domain

        Parameters
        ------------

        site_type : string
            The site type identifier for which the function will search 
            for matching sites

        return_list : bool
            By default, the flag returns a dictionary, which is convenient as 
            it makes it easy to index into one or more sites at a specific 
            position in the sequence. However, you may instead want a list 
            of sites, in which case setting return_list will have the function
            simply return a list of sites. As of right now we do not guarentee
            the order of these returned sites.

        
        Returns
        --------

        list
            Returns a dictionary, where each key-value pair is:

                key - site position (integer)
                value - list of one or more site object
        
        """

        return self._protein.get_sites_by_type_and_range(site_type, self.start, self.end)
        

    #######################################
    ##                                   ##
    ##      DOMAIN TRACK FUNCTIONS       ##
    ##                                   ##
    #######################################

    ## ------------------------------------------------------------------------
    ##      
    def get_track_values(self, name, safe=True):
        """
        Function that returns the region of a protein's values- track 
        associated with this domain.        
        
        If the track name is not found in this protein and safe is True, 
        this will throw an exception, otherwise (if safe=False) then if 
        the track is missing the function will return None.

        Parameters
        --------------
        
        name : str
            Track name

        safe : bool (default = True)
            If set to True, missing tracks trigger an exception, else 
            they just return None

        Returns
        ----------
        list
            Returns a list of floats that corresponds to the set of 
            residues associated with the domain of interest, or None if 
            the track does not exist and safe=False.
            

        """
        
        t = self._protein.track(name, safe)

        if t is not None:
            try:
                return t.values_region(self._start, self._end)
            except TypeError:
                if t.values == None:
                    raise DomainException('Passed associated track has no values - try get_track_symbols')
                else:
                    raise DomainException('Error with passed associated track')
        else:
            return None


    ## ------------------------------------------------------------------------
    ##      
    def get_track_symbols(self, name, safe=True):
        """
        Function that returns the region of a protein's symbols track 
        associated with this domain.
                
        If the track name is missing and safe is True, this will throw 
        an exception, otherwise (if safe=False) then if the track is 
        missing the function returns None
        
        Parameters
        --------------
        
        name : str
            Track name

        safe : bool (default = True)
            If set to True, missing tracks trigger an exception, else 
            they just return None

        Returns
        ----------
        list
            Returns a list of strings that corresponds to the set of 
            residues associated with the domain of interest.

        """

        t = self._protein.track(name, safe)
        
        if t is not None:
            try:
                return t.symbols_region(self._start, self._end)
            except TypeError:
                if t.symbols == None:
                    raise DomainException('Passed associated track has no symbols - try get_track_values')
                else:
                    raise DomainException('Error with passed associated track')
        else:
            return None
        

    ## ------------------------------------------------------------------------
    ##      
    def __repr__(self):
        return "|Domain: %s (%i-%i, len=%i) in protein %s" % (self._domain_type, self.start, self.end, len(self), self.protein.unique_ID)

    ## ------------------------------------------------------------------------
    ##      
    def __len__(self):
        return len(self.sequence)

    
            






"""
SHEPHARD: 
Sequence-based Hierarchical and Extendable Platform for High-throughput Analysis of Region of Disorder

Authors: Garrett M. Ginell & Alex S. Holehouse
Contact: (g.ginell@wustl.edu)

Holehouse Lab - Washington University in St. Louis
"""

"""
Empty init file in case you choose a package besides PyTest such as Nose which may look for such a file
"""



"""
SHEPHARD: 
Sequence-based Hierarchical and Extendable Platform for High-throughput Analysis of Region of Disorder

Authors: Garrett M. Ginell & Alex S. Holehouse
Contact: (g.ginell@wustl.edu)

Holehouse Lab - Washington University in St. Louis
"""
import re

def inside_region(boundary_start, boundary_end, position):
    """
    Given a position, does the position fall inside the defined region, inclusive
    for both start and end positions
    """
    if position >= boundary_start and position <= boundary_end:
        return True
    else:
        return False



def update_from_realworld(vals, use_realworld_numbering=True):
    """
    Takes in a list of values and IF use_realworld_numbering is
    true will subtract 1 from each (i.e. moving from a 1-indexed
    numbering to a 0-indexed numbering).

    If use_realword is false then the same numbers are returned
    and no alterations are made.

    """

    if use_realworld_numbering:
        return_vals = []
        for i in vals:
            return_vals.append(i - 1)
        return return_vals

    else:
        return vals


def update_to_realworld(vals, use_realworld_numbering=True):
    """
    Takes in a list of values and IF use_realworld_numbering is
    true will add 1 from each (i.e. moving from a 0-indexed
    numbering to a 1-indexed numbering).

    If use_realword is false then the same numbers are returned
    and no alterations are made.

    """

    if use_realworld_numbering:
        return_vals = []
        for i in vals:
            return_vals.append(i + 1)
        return tuple(return_vals)

    else:
        return tuple(vals)



def get_bounding_sites(position, offset, maxlen):
    # compute start/end of context according to the offset
    p1 = max(1, position - offset)
    p2 = min(maxlen, position + offset)
    return (p1,p2)

            



"""
SHEPHARD: 
Sequence-based Hierarchical and Extendable Platform for High-throughput Analysis of Region of Disorder

Authors: Garrett M. Ginell & Alex S. Holehouse
Contact: (g.ginell@wustl.edu)

Holehouse Lab - Washington University in St. Louis
"""

# must import pytest
import pytest

# import additional modules needed for this code
import shephard
from shephard.apis import uniprot  
from shephard import interfaces

# import the domain_tools module which we're going to test
from shephard.tools import domain_tools




"""
SHEPHARD: 
Sequence-based Hierarchical and Extendable Platform for High-throughput Analysis of Region of Disorder

Authors: Garrett M. Ginell & Alex S. Holehouse
Contact: (g.ginell@wustl.edu)

Holehouse Lab - Washington University in St. Louis
"""

import protfasta
from shephard.proteome import Proteome
from shephard.exceptions import APIException

SHEPHARD_ATTRIBUTE_SPLITTER='SHPRD_ATTRIBUTES='

## ------------------------------------------------------------------------
##
def fasta_to_proteome(filename, 
                      proteome=None, 
                      build_unique_ID=None, 
                      build_attributes=None, 
                      use_header_as_unique_ID=False, 
                      force_overwrite=False,
                      invalid_sequence_action='fail'):
    """
    Stand alone function that allows the user to build a Proteome from a 
    standard FASTA file, or add sequences in a FASTA file to an existing 
    Proteome.

    This function can be used to read additional sequences into an existing
    Proteome object, or create a new Proteome object from a FASTA file. 
    In addition, some control over how invalid sequences should be dealt 
    with are defined by the invalid_sequence_action flag.
    
    The input filename must be a FASTA file without duplicate headers. If 
    the file has duplicate headers and these have to be further processed 
    we suggest using the protfasta (https://protfasta.readthedocs.io/) 
    package to parse through the FASTA file first creating a santizied 
    input FASTA.

    The FASTA file is parsed into a set of proteins, each of which has (1) 
    a unique ID, (2) a name, (3) a sequence, and (4) optionally, a 
    dictionary of attributes.

    The protein name is defined as the full FASTA header, and the 
    sequence based on the FASTA record sequence. Sequence validation is
    also provided at the point of file-parsing. The unique ID and 
    attributes are discused below.
            
    Each protein in a Proteome must have a unique_ID associated with it. 
    There are three ways a FASTA file can be used to generate a unique ID:

        1. By using the FASTA header as a unique ID, although this fails
           if there are non-unique FASTA headers.

        2. By parsing the FASTA header, to extract out a unique ID. For 
           example, FASTA files generated by other databases often include
           unique identifiers in a structured way, which could be extracted
           in a consistent manner for every FASTA record.

        3. By incrementing an automatically unique ID, removing any 
           dependence on the FASTA header itself.

    These options can be selected using the flags provided in the function 
    signature. Note that if both using the FASTA header directly and parsing
    the FASTA header are selected an exception will be raised as only one
    of these two can be requested simultaneously.

    By default, a numerically unique value is used. Note that if your are
    reading FASTA files generated by UniProt, we recommend the api.uniprot
    functions instead of the more generic api.fasta functions.

    To build protein attributes, this can in principle be parsed out of the 
    FASTA header by providing a build_attributes() function. However, we 
    would in general suggest that the better way to do this would be annotate
    a Proteome with attribibutes and the same those using the associated
    interfaces.si_protein_attributes functionality.    
    
    
    Parameters
    ------------

    filename : string
        Name of the FASTA file we're going to parse in. 

    proteome : Proteome (default = None)
        If a Proteome object is provided the FASTA file will be read and 
        added to the existing proteome, whereas if set to None a new 
        Proteome will be generated.        

    build_unique_ID : funct (default = None)
        This parameter allows a user-defined function that is used to 
        convert the FASTA header to a (hopefully) unique string. This can 
        be useful if the FASTA header is well structured and includes 
        a specific, useful unique ID that can be used as the unique_ID.

        Specifically, the build_unique_ID function should take in the 
        a str (a FASTA header) and return a string which will be used
        as a unique ID.
        
    build_attributes : funct (default = None)
        This parameter allows a user-defined function that allows 
        meta-information from the FASTA header to be converted into 
        protein attributes. Specifically, build_attributes should be a 
        function which takes in the FASTA header as a string and returns 
        a dictionary where key:value pairs are assigned as protein attributes. 
        This can be useful if the FASTA header is well-structured. 
    
    use_header_as_unique_ID : bool (default = False)
        If this flag is set to true, it means the unique_ID is set to the FASTA 
        file header. If non-unique headers are found this will trigger an 
        exception.

    force_overwrite : bool (default  = False)
        If this flag is set to true  and we encounter a unique_ID that is 
        already in the proteome the newer value overwrites the older one. 
        This is mostly useful if you are adding in a file with known 
        duplicate entries OR combining multiple FASTA files where you know 
        there's some duplications. Important - if we're building unique IDs
        based on numerical record indices then EVERY FASTA entry will be given 
        a unique_ID (meaning force_overwrite is irrelevant in this case).

    invalid_sequence_action : str (default = 'fail')
        Selector which defines the behaviour if a sequence with a non-
        standard amino acid is encountered. Valid options and their meaning
        are listed below:

            * ``ignore``  - invalid sequences are completely ignored.

            * ``fail``    - invalid sequence cause parsing to fail and throw an exception.
                            
            * ``remove`` -  invalid sequences are removed.

            * ``convert`` - invalid residues are converted to valid residues.
                            
            * ``convert-ignore`` - invalid sequences are converted to valid sequences and any remaining invalid residues are ignored.

        
    Returns 
    --------
    Proteome
        Returns an initialized Proteome object 
    
    """

    # parameter sanity checking
    if use_header_as_unique_ID is True and build_unique_ID is not None:
        raise APIException('Cannot simultaneously set use_header_as_unique_ID = True and build_unique_ID to not None')
        
    # read in the fasta file using protfasta
    fasta_dictionary = protfasta.read_fasta(filename, invalid_sequence_action=invalid_sequence_action)

    # extract the keys (FASTA headers) and initialize the record_index (internal
    # numbering used for construction. Also initialize the proteom_dict, which is
    # a dictionary of protein entries we passed to Proteome.
    record_index  = 0

    # IF we're adding to a new proteome this bit of code sets the record_index to the largest new integer
    # such that we can add multiple proteomes in succession and we'll get a proteome where there are numerically
    # contigous unique_IDs.  Note we only do this if we'll be using the record_index
    if proteome is not None and (build_unique_ID is None or use_header_as_unique_ID is None):
        numeric_record_ids = []
        for uid in proteome.proteins:
            try:
                numeric_record_ids.append(int(uid))
            except ValueError:
                pass
        if len(numeric_record_ids) > 0:
            record_index = max(numeric_record_ids)+1

    # initialize the empty list
    proteome_list = []

    # for each entry
    for k in fasta_dictionary:

        # create a key-value pair where 
        #   key = the unique record_index (this is only used for internal structure
        #         within this function to assure we never overwrite in this dictionary
        #
        #  value = a four-position list where the positions reflect the following
        #        [0] = amino acid sequence
        #        [1] = name (this can be anything)
        #        [2] = unique_ID - this should be a unique identifier that can be used
        #              to cross-reference this entry to other data. If extrat_unique_ID
        #              is passed we try to use this 
        #        [3] = attribute dictionary (we set this to None)
        
        
        # get unique_ID 
        if build_unique_ID:
            unique_ID = build_unique_ID(k)
        elif use_header_as_unique_ID is True:
            unique_ID = k
        else:
            unique_ID = record_index
        
        # build an attributes dictionary using the user-provided custom function
        if build_attributes:
            attributes = build_attributes(k)
        else:
            attributes = {}
            
        # now create an input dictionary orbject
        newdict = {}
        newdict['sequence'] = str(fasta_dictionary[k])
        newdict['name'] = k
        newdict['unique_ID'] = unique_ID
        newdict['attributes'] = attributes

        proteome_list.append(newdict)

        record_index = record_index + 1
        
    # finally if a proteome was provided then 
    if proteome is not None:
        proteome.add_proteins(proteome_list, force_overwrite=force_overwrite)
        return proteome
    else:
        # no proteome provided so build a new proteome and return it    
        return Proteome(proteome_list, force_overwrite=force_overwrite)



## ------------------------------------------------------------------------
##
def shephard_fasta_to_proteome(filename, 
                              proteome = None,
                              force_overwrite=False,
                              invalid_sequence_action='fail'):
                              
    """
    Stand alone function that allows the user to build a proteome 
    from a FASTA file generated by SHEPHARD (using the proteome_to_fasta() 
    function. When SHEPHARD generates a FASTA file it uses a general 
    convention for encoding the unique ID, protein name, and attributes.

    Specifically, the FASTA header has the form 

    >SHPRD|<UNIQUE_ID>|<PROTEIN_NAME|SHPRD_ATTRIBUTES=<ATTRIBUTE_NAME>=<ATTRIBUTE_VALUE>\t

    Where an arbitrary number of name/value attribute paris can be encoded
    separated by a tab character.

    WARNING: The support for protein attributes in FASTA files is included
    mainly for easy sharing of FASTA files that are usable outside of 
    SHEPHARD. We recommend using interfaces.si_protein_attributes functions
    for dealing with Protein attributes.


    Parameters
    ------------

    filename : str
        Name of the FASTA file we're going to parse in. 
        
    proteome : Proteome (default = None)
        If a Proteome object is provided the FASTA file will be read and 
        added to the existing proteome, whereas if set to None a new 
        Proteome will be generated.

    force_overwrite : bool (default  = False)
        If this flag is set to true  and we encounter a unique_ID that is 
        already in the proteome the newer value overwrites the older one. 
        This is mostly useful if you are adding in a file with known 
        duplicate entries OR combining multiple FASTA files where you know 
        there's some duplications. Important - if we're building unique IDs
        based on numerical record indices then EVERY FASTA entry will be given 
        a unique_ID (meaning force_overwrite is irrelevant in this case).

    invalid_sequence_action : str (default = 'fail')
        Selector which defines the behaviour if a sequence with a non-
        standard amino acid is encountered. Valid options and their meaning
        are listed below:

            * ``ignore``  - invalid sequences are completely ignored.

            * ``fail``    - invalid sequence cause parsing to fail and throw an exception.
                            
            * ``remove`` -  invalid sequences are removed.

            * ``convert`` - invalid residues are converted to valid residues.
                            
            * ``convert-ignore`` - invalid sequences are converted to valid sequences and any remaining invalid residues are ignored.
                                   
                                   
    Returns 
    --------
    Proteome Object
        Returns an initialized Proteome object 
    
    """
        
    # read in the fasta file using protfasta
    fasta_dictionary = protfasta.read_fasta(filename, invalid_sequence_action=invalid_sequence_action)

    # initialize the empty list
    proteome_list = []

    # for each entry    
    for k in fasta_dictionary:
        
        # because we know what the header format will be we can be definitive about extracting the relevant information
        fasta_split = k.split('|')

        # ENSURE EVERY single line is a valid 
        if fasta_split[0] != "SHPRD":
            raise APIException('Trying to parse a FASTA file that is expected to be SHEPHARD generated but formatting does not comply [on entry %s in file %s]' % (k, filename))
        
        # extract out 
        try:
            # get the unique ID
            unique_ID = fasta_split[1]

            # then take everything after the unique_ID
            tmp = "|".join(fasta_split[2:])
            attributes_string = tmp.split(SHEPHARD_ATTRIBUTE_SPLITTER)
            name = attributes_string[0]
        except IndexError:
            raise APIException('Trying to parse a FASTA file that is expected to be SHEPHARD generated but formatting does not comply [on entry %s in file %s]' % (k, filename))

        attributes_dict = {}

        if len(attributes_string) > 1:            
            attributes_string_s = attributes_string[1].split('\t')

            for a in attributes_string_s:
                local_k = a.strip().split('=')[0].strip()
                local_v = a.strip().split('=')[0].strip()
                attributes_dict[local_k] = local_v
                                        
        # now create an protein dictionary object and populate!
        newdict = {}
        newdict['sequence'] = str(fasta_dictionary[k])
        newdict['name'] = name
        newdict['unique_ID'] = unique_ID
        newdict['attributes'] = attributes_dict

        proteome_list.append(newdict)
        
    # finally if a proteome was provided then 
    if proteome is not None:
        proteome.add_proteins(proteome_list, force_overwrite=force_overwrite)
        return proteome
    else:
        # no proteome provided so build a new proteome and return it    
        return Proteome(proteome_list, force_overwrite=force_overwrite)


## ------------------------------------------------------------------------
##
def proteome_to_fasta(filename, proteome, include_attributes_in_header=False):
    """
    Stand alone function that allows the user to write a SHEPHARD-specific 
    FASTA file from a Proteome object.

    WARNING: The support for protein attributes in FASTA files is included
    mainly for easy sharing of FASTA files that are usable outside of 
    SHEPHARD. We recommend using interfaces.si_protein_attributes functions
    for dealing with Protein attributes.

    Parameters
    ------------
    filename : str
        Name of the FASTA file we're going to write to. We will automatically 
        overwrite a file if it's there, so be careful! Note that no extension 
        is added in part because FASTA files can be .f/.fa/.fasta. 
        Recommended a .fasta file extension.

    proteome : Proteome
        The proteome object that will be written to disk
        
    include_attributes_in_header : bool (default = falseFalse)
        Flag which if set to true means each Protein's attributes will be 
        included in the FASTA header. We generally do not recommend this 
        other than times when sharing annotated FASTA files outside of a 
        SHEPHARD ecosystem would be useful.

    Returns
    -----------
    None
        No return object but a new file will be written

    """

    # build output list with or without the unique_ID 
    outlist = []
    for protein in proteome:

        # this is where we define the FASTA header...
        fasta_header = "SHPRD|%s|%s" % (protein.unique_ID, protein.name)
            
        # this is where we append the FASTA header with attributes    
        if include_attributes_in_header:

            fasta_header = fasta_header + " " + SHEPHARD_ATTRIBUTE_SPLITTER
            
            for k in protein.attributes:
                
                # these lines ensure there are no tabs in the attribute names of
                # values before we append them to the header file, ensuring that
                # IF we want to read the fasta header info back in to attributes
                # we can be confident that hidden tabs in the variables won't
                # mess things up!
                k_fixed = k.replace('\t', ' ' )
                i = protein.attribute(k)
                i_fixed = i.replace('\t', ' ')
                
                fasta_header = fasta_header + '\t' + "%s=%s" %(k_fixed, i_fixed)
            
        outlist.append([fasta_header, protein.sequence])
        
    # use the protfasta library to write the file to disk
    protfasta.write_fasta(outlist, filename, linelength=80)
                





import pytest
import shephard
from shephard import proteome
from shephard.apis import fasta
from shephard.exceptions import ProteomeException, APIException


def test_fasta_to_proteome_part_1():

    test_data_dir = shephard.get_data('test_data')
    print(test_data_dir)

    P = fasta.fasta_to_proteome('%s/%s' % (test_data_dir, 'testset_1.fasta'))
    assert len(P.protein('1')) == 390
    assert len(P) == 9

    test_UID = 0
    for i in P.proteins:
        assert i == str(test_UID)
        test_UID = test_UID + 1

    P = fasta.fasta_to_proteome('%s/%s' % (test_data_dir, 'testset_1.fasta'), proteome=P)
    assert len(P) == 18

    test_UID = 0
    for i in P.proteins:
        assert i == str(test_UID)
        test_UID = test_UID + 1


    ##
    ## This block checks that removing a protein from the integer-indexed added proteins
    ## really removes it and that adding new proteins in does correctly start counting 
    ## in the right place
    P.remove_protein(10)
    P = fasta.fasta_to_proteome('%s/%s' % (test_data_dir, 'testset_1.fasta'), proteome=P)
        
    test_UID = 0
    for i in P.proteins:
        if test_UID == 10:
            with pytest.raises(ProteomeException):
                assert P.protein(test_UID)
            test_UID = test_UID + 1

        assert i == str(test_UID)
        test_UID = test_UID + 1

    
def test_fasta_to_proteome_part_2():
    def header_parser(s):
        return s.split('|')[1]

    test_data_dir = shephard.get_data('test_data')
    print(test_data_dir)

    # check we've read in the proteome OK
    P = fasta.fasta_to_proteome('%s/%s' % (test_data_dir, 'testset_1.fasta'), build_unique_ID=header_parser)
    assert len(P) == 9

    # this should trigger an exception because we're adding in duplicates
    with pytest.raises(ProteomeException):
        P = fasta.fasta_to_proteome('%s/%s' % (test_data_dir, 'testset_1.fasta'), build_unique_ID=header_parser, proteome=P)

    # this should not...
    P = fasta.fasta_to_proteome('%s/%s' % (test_data_dir, 'testset_1.fasta'), proteome=P)
    assert len(P) == 18

    # this also should not but should NOT add new sequences in
    P = fasta.fasta_to_proteome('%s/%s' % (test_data_dir, 'testset_1.fasta'), build_unique_ID=header_parser, proteome=P, force_overwrite=True)
    assert len(P) == 18

    P = fasta.fasta_to_proteome('%s/%s' % (test_data_dir, 'testset_1.fasta'), proteome=P)
    assert len(P) == 27

    P = fasta.fasta_to_proteome('%s/%s' % (test_data_dir, 'testset_1.fasta'), proteome=P, use_header_as_unique_ID=True)
    assert len(P) == 36

    # this SHOULD trigger an exception because we shouldn't
    with pytest.raises(APIException):
        P = fasta.fasta_to_proteome('%s/%s' % (test_data_dir, 'testset_1.fasta'), proteome=P, use_header_as_unique_ID=True, build_unique_ID=header_parser)

    expected_protein_uids = ['O00401', 'O00470', 'O00472', 'O00499', 'O00629', 'O00712', 'O00716', 'O14786', 'Q9UJX3', '0', '1', '2', '3', '4', '5', '6', '7', '8', '9', '10', '11', '12', '13', '14', '15', '16', '17', 'sp|O00401|WASL_HUMAN Neural Wiskott-Aldrich syndrome protein OS=Homo sapiens OX=9606 GN=WASL PE=1 SV=2', 'sp|O00470|MEIS1_HUMAN Homeobox protein Meis1 OS=Homo sapiens OX=9606 GN=MEIS1 PE=1 SV=1', 'sp|O00472|ELL2_HUMAN RNA polymerase II elongation factor ELL2 OS=Homo sapiens OX=9606 GN=ELL2 PE=1 SV=2', 'sp|O00499|BIN1_HUMAN Myc box-dependent-interacting protein 1 OS=Homo sapiens OX=9606 GN=BIN1 PE=1 SV=1', 'sp|O00629|IMA3_HUMAN Importin subunit alpha-3 OS=Homo sapiens OX=9606 GN=KPNA4 PE=1 SV=1', 'sp|O00712|NFIB_HUMAN Nuclear factor 1 B-type OS=Homo sapiens OX=9606 GN=NFIB PE=1 SV=2', 'sp|O00716|E2F3_HUMAN Transcription factor E2F3 OS=Homo sapiens OX=9606 GN=E2F3 PE=1 SV=1', 'sp|O14786|NRP1_HUMAN Neuropilin-1 OS=Homo sapiens OX=9606 GN=NRP1 PE=1 SV=3', 'sp|Q9UJX3|APC7_HUMAN Anaphase-promoting complex subunit 7 OS=Homo sapiens OX=9606 GN=ANAPC7 PE=1 SV=4']

    # CHECK all the unique IDs expected can be read in
    for i in expected_protein_uids:
        P.protein(i) 

    

    


    
    
    



"""
SHEPHARD: 
Sequence-based Hierarchical and Extendable Platform for High-throughput Analysis of Region of Disorder

Authors: Garrett M. Ginell & Alex S. Holehouse
Contact: (g.ginell@wustl.edu)

Holehouse Lab - Washington University in St. Louis
"""

# Import package, test suite, and other packages as needed
import shephard
import pytest
import sys

def test_shephard_imported():
    """Sample test, will always pass so long as import statement worked"""
    assert "shephard" in sys.modules



import pytest

import shephard
from shephard.apis import uniprot  
from shephard import interfaces
import numpy as np
from shephard.exceptions import ProteinException, InterfaceException
import os
import random



# ....................................................................................................
#
#
def test_add_track_from_file():

    test_data_dir = shephard.get_data('test_data')
    fasta_file = '%s/%s' % (test_data_dir, 'testset_1.fasta')
    track_file = '%s/%s' % (test_data_dir, 'TS1_tracks_pscore.tsv')
    
    
    def testfunct(P):
        assert P.protein('O00401').track('pscore').values_region(50,50)[0] == -0.971
        assert P.protein('O00470').track('pscore').values_region(50,50)[0] == 0.324
        assert P.protein('O00472').track('pscore').values_region(50,50)[0] == -0.661
        assert P.protein('O00499').track('pscore').values_region(50,50)[0] == 0.205

    
        assert np.sum(P.protein('O00499').track('pscore').values_region(500,593)) == 43.647999999999996

        with pytest.raises(ProteinException):
            assert 1 == P.protein('O00499').track('pscore').values_region(594, 594)[0]

        
    # test various permutations of reading in (all should work)

    P = uniprot.uniprot_fasta_to_proteome(fasta_file)
    interfaces.si_tracks.add_tracks_from_file(P, track_file, mode='values', safe=False, verbose=True, skip_bad=False)
    testfunct(P)

    P = uniprot.uniprot_fasta_to_proteome(fasta_file)
    interfaces.si_tracks.add_tracks_from_file(P, track_file, mode='values', verbose=True, skip_bad=False)
    testfunct(P)

    P = uniprot.uniprot_fasta_to_proteome(fasta_file)
    interfaces.si_tracks.add_tracks_from_file(P, track_file, mode='values', safe=False, skip_bad=False)
    testfunct(P)

    P = uniprot.uniprot_fasta_to_proteome(fasta_file)
    interfaces.si_tracks.add_tracks_from_file(P, track_file, mode='values', safe=False, skip_bad=True)
    testfunct(P)

    P = uniprot.uniprot_fasta_to_proteome(fasta_file)
    interfaces.si_tracks.add_tracks_from_file(P, track_file, mode='values', safe=True, skip_bad=True)
    testfunct(P)

    P = uniprot.uniprot_fasta_to_proteome(fasta_file)
    interfaces.si_tracks.add_tracks_from_file(P, track_file, mode='values', safe=False, verbose=True)
    testfunct(P)

    P = uniprot.uniprot_fasta_to_proteome(fasta_file)
    interfaces.si_tracks.add_tracks_from_file(P, track_file, mode='values', verbose=True)
    testfunct(P)

    P = uniprot.uniprot_fasta_to_proteome(fasta_file)
    interfaces.si_tracks.add_tracks_from_file(P, track_file, mode='values', skip_bad=True)
    testfunct(P)

    P = uniprot.uniprot_fasta_to_proteome(fasta_file)
    interfaces.si_tracks.add_tracks_from_file(P, track_file, mode='values', safe=False)
    testfunct(P)


# ....................................................................................................
#
#
def test_add_track_from_dictionary():

    test_data_dir = shephard.get_data('test_data')
    fasta_file = '%s/%s' % (test_data_dir, 'testset_1.fasta')
    
    P = uniprot.uniprot_fasta_to_proteome(fasta_file)

    
    random_vals = np.random.rand(len(P.protein('O00401')))
    track_dict = {'O00401': [{'track_name':'random_vals', 'track_data':random_vals}]}
    interfaces.si_tracks.add_tracks_from_dictionary(P, track_dict, mode='values')
    assert len(P.protein('O00401')) == len(P.protein('O00401').track('random_vals').values)

    newstring = ''
    for i in P.protein('O00401').sequence:

        if i in ['S','T','Y']:
            newstring = newstring + "P"
        else:
            newstring = newstring + "-"

    track_dict = {'O00401': [{'track_name':'phosres', 'track_data':newstring}]}
    interfaces.si_tracks.add_tracks_from_dictionary(P, track_dict, mode='symbols')


    with pytest.raises(TypeError):
        assert len(P.protein('O00401')) == len(P.protein('O00401').track('phosres').values)

    assert "".join(P.protein('O00401').track('phosres').symbols_region(1,10)) == "-PP-------"
    
    with pytest.raises(ProteinException):
        assert "".join(P.protein('O00401').track('phosres').symbols_region(0,10)) == "-PP-------"

    with pytest.raises(ProteinException):
        local_p = P.protein('O00401')
        print(local_p.track('phosres').symbols_region(len(local_p), len(local_p)+10))



# ....................................................................................................
#
#
def test_write_track_part1():

    test_data_dir = shephard.get_data('test_data')
    fasta_file = '%s/%s' % (test_data_dir, 'testset_1.fasta')
    track_file = '%s/%s' % (test_data_dir, 'TS1_tracks_pscore.tsv')
    
    # write
    P = uniprot.uniprot_fasta_to_proteome(fasta_file)
    interfaces.si_tracks.add_tracks_from_file(P, track_file, mode='values')

    fn = 'output_test/out_tracks.tsv'
    interfaces.si_tracks.write_tracks(P, fn, 'pscore')

    b = os.path.getsize(fn)
    assert b == 32531 # file size in bytes if correctly written

    P2 = uniprot.uniprot_fasta_to_proteome(fasta_file)
    interfaces.si_tracks.add_tracks_from_file(P2, 'output_test/out_tracks.tsv', mode='values')

    for idx in P.proteins:
        protein = P.protein(idx)
        protein2 = P2.protein(idx)
        
        assert np.sum(protein.track('pscore').values) == np.sum(protein2.track('pscore').values)

# ....................................................................................................
#
#
def test_write_track_part2():

    test_data_dir = shephard.get_data('test_data')
    fasta_file = '%s/%s' % (test_data_dir, 'testset_1.fasta')
    pscore_track_file = '%s/%s' % (test_data_dir, 'TS1_tracks_pscore.tsv')


    # build proteome
    P = uniprot.uniprot_fasta_to_proteome(fasta_file)

    # create an additional values track
    special_ID = 'O00401'
    random_vals = np.random.rand(len(P.protein(special_ID)))
    track_name = 'random_vals'
    outname = 'output_test/track_%s.tsv'%(track_name)

    track_dict = {special_ID: [{'track_name':track_name, 'track_data':random_vals}]}
    interfaces.si_tracks.add_tracks_from_dictionary(P, track_dict, mode='values')
    interfaces.si_tracks.write_tracks(P, outname,track_name)

    b = os.path.getsize(outname)
    assert b == 3050 # file size in bytes if correctly written

    P2 = uniprot.uniprot_fasta_to_proteome(fasta_file)
    interfaces.si_tracks.add_tracks_from_file(P2, pscore_track_file, mode='values')
    interfaces.si_tracks.add_tracks_from_file(P2, outname, mode='values')

    for protein in P2:
        if protein.unique_ID == special_ID:
            assert len(protein.tracks) == 2                
        else:
            assert len(protein.tracks) == 1       


def test_write_track_part3():

    test_data_dir = shephard.get_data('test_data')
    fasta_file = '%s/%s' % (test_data_dir, 'testset_1.fasta')
    pscore_track_file = '%s/%s' % (test_data_dir, 'TS1_tracks_pscore.tsv')


    P = uniprot.uniprot_fasta_to_proteome(fasta_file)
    interfaces.si_tracks.add_tracks_from_file(P, pscore_track_file, mode='values')

    # ..................................................................
    ## First create and write a random vals track
    # create an additional values track
    special_ID = 'O00401'
    random_vals = np.random.rand(len(P.protein(special_ID)))
    track_name_rv = 'random_vals'
    outname_rv = 'output_test/track_%s.tsv'%(track_name_rv)

    track_dict = {special_ID: [{'track_name':track_name_rv, 'track_data':random_vals}]}
    interfaces.si_tracks.add_tracks_from_dictionary(P, track_dict, mode='values')
    interfaces.si_tracks.write_tracks(P, outname_rv, track_name_rv)


    track_name = 'phospho_track'
    outname = 'output_test/track_%s.tsv'%(track_name)


    # create an additional symbols tracks and write track
    newstring = ''
    for i in P.protein(special_ID).sequence:

        if i in ['S','T','Y']:
            newstring = newstring + "P"
        else:
            newstring = newstring + "-"

    track_dict = {special_ID: [{'track_name':track_name, 'track_data':newstring}]}
    interfaces.si_tracks.add_tracks_from_dictionary(P, track_dict, mode='symbols')    
    interfaces.si_tracks.write_tracks(P, outname, track_name)

    b = os.path.getsize(outname)
    assert b == 1032 # file size in bytes if correctly written

    P3 = uniprot.uniprot_fasta_to_proteome(fasta_file)
    interfaces.si_tracks.add_tracks_from_file(P3, outname, mode='symbols')
    interfaces.si_tracks.add_tracks_from_file(P3, outname_rv, mode='values')
    
    for protein in P3:
        if protein.unique_ID == special_ID:
            assert len(protein.tracks) == 2               
        else:
            assert len(protein.tracks) == 0      


def test_write_tracks_separate_files():

    ##
    ## This code block just creates a proteome where we have 2 different
    ## typs of tracks
    test_data_dir = shephard.get_data('test_data')
    fasta_file = '%s/%s' % (test_data_dir, 'testset_1.fasta')
    pscore_track_file = '%s/%s' % (test_data_dir, 'TS1_tracks_pscore.tsv')

    P = uniprot.uniprot_fasta_to_proteome(fasta_file)
    interfaces.si_tracks.add_tracks_from_file(P, pscore_track_file, mode='values')

    newstring = ''
    for i in P.protein('O00401').sequence:

        if i in ['S','T','Y']:
            newstring = newstring + "P"
        else:
            newstring = newstring + "-"
    track_dict = {'O00401': [{'track_name':'phosres', 'track_data':newstring}]}
    interfaces.si_tracks.add_tracks_from_dictionary(P, track_dict, mode='symbols')

    interfaces.si_tracks.write_all_tracks_separate_files(P, 'output_test')



def test_write_tracks_separate_files():

    ##
    ## This code block just creates a proteome where we have 2 different
    ## typs of tracks
    test_data_dir = shephard.get_data('test_data')
    fasta_file = '%s/%s' % (test_data_dir, 'testset_1.fasta')
    pscore_track_file = '%s/%s' % (test_data_dir, 'TS1_tracks_pscore.tsv')

    P = uniprot.uniprot_fasta_to_proteome(fasta_file)
    interfaces.si_tracks.add_tracks_from_file(P, pscore_track_file, mode='values')

    newstring = ''
    for i in P.protein('O00401').sequence:

        if i in ['S','T','Y']:
            newstring = newstring + "P"
        else:
            newstring = newstring + "-"
    track_dict = {'O00401': [{'track_name':'phosres', 'track_data':newstring}]}
    interfaces.si_tracks.add_tracks_from_dictionary(P, track_dict, mode='symbols')

    interfaces.si_tracks.write_all_tracks_separate_files(P, 'output_test')

    P2 = uniprot.uniprot_fasta_to_proteome(fasta_file)
    interfaces.si_tracks.add_tracks_from_file(P2, 'output_test/shephard_track_pscore.tsv', mode='values')

    for idx in P2.proteins:
        assert P2.protein(idx).track('pscore').values == P.protein(idx).track('pscore').values


    P2 = uniprot.uniprot_fasta_to_proteome(fasta_file)
    interfaces.si_tracks.add_tracks_from_file(P2, 'output_test/shephard_track_phosres.tsv', mode='symbols')

    for idx in ['O00401']:
        for position in range(1,len(P2.protein(idx))+1):
            POS1 = P.protein(idx).track('phosres').symbols_region(position,position)
            POS2 = P2.protein(idx).track('phosres').symbols_region(position,position)
            assert POS1 == POS2

        
def test_write_tracks_single_file_symbols():

    test_data_dir = shephard.get_data('test_data')
    fasta_file = '%s/%s' % (test_data_dir, 'testset_1.fasta')
    
    P = uniprot.uniprot_fasta_to_proteome(fasta_file)

    # build phosphostring
    newstring = ''
    for i in P.protein('O00401').sequence:

        if i in ['S','T','Y']:
            newstring = newstring + "P"
        else:
            newstring = newstring + "-"
    track_dict = {'O00401': [{'track_name':'phosres', 'track_data':newstring}]}
    interfaces.si_tracks.add_tracks_from_dictionary(P, track_dict, mode='symbols')

    # build phosphostring
    newstring = ''
    for i in P.protein('O00472').sequence:

        if i in ['Y','F','W']:
            newstring = newstring + "A"
        else:
            newstring = newstring + "-"
    track_dict = {'O00472': [{'track_name':'aro', 'track_data':newstring}]}
    
    interfaces.si_tracks.add_tracks_from_dictionary(P, track_dict, mode='symbols')
    interfaces.si_tracks.write_all_symbols_tracks_single_file(P, 'output_test/all_symbols_test.tsv')
    
    P2 = uniprot.uniprot_fasta_to_proteome(fasta_file)
    interfaces.si_tracks.add_tracks_from_file(P2, 'output_test/all_symbols_test.tsv', mode='symbols')

    idx = 'O00401'
    for position in range(1,len(P2.protein(idx))+1):
        POS1 = P.protein(idx).track('phosres').symbols_region(position,position)
        POS2 = P2.protein(idx).track('phosres').symbols_region(position,position)
        assert POS1 == POS2


    idx = 'O00472'
    for position in range(1,len(P2.protein(idx))+1):
        POS1 = P.protein(idx).track('aro').symbols_region(position,position)
        POS2 = P2.protein(idx).track('aro').symbols_region(position,position)
        assert POS1 == POS2


def test_write_tracks_single_file_values():

    test_data_dir = shephard.get_data('test_data')
    fasta_file = '%s/%s' % (test_data_dir, 'testset_1.fasta')
    
    P = uniprot.uniprot_fasta_to_proteome(fasta_file)

    # build phosphostring

    r1 = random.randint(1,20)
    r2 = random.random()
    for protein in P:

        new_vals = []
        for i in protein.sequence:

            if i in ['S','T','Y']:
                new_vals.append(r1)
            else:
                new_vals.append(r2)

        protein.add_track('test_track', values=new_vals)

    # write all the tracks out
    interfaces.si_tracks.write_all_values_tracks_single_file(P, 'output_test/all_values_test.tsv')
    
    # read the tracks into a new proteome
    P2 = uniprot.uniprot_fasta_to_proteome(fasta_file)
    interfaces.si_tracks.add_tracks_from_file(P2, 'output_test/all_values_test.tsv', mode='values')

    # check everything worked ok
    for protein in P2:
        print(protein)
        for i in range(1,len(protein)+1):

            if protein.residue(i) in ['S','T','Y']:
                assert protein.track('test_track').value(i) == r1
            else:
                assert protein.track('test_track').value(i) == np.round(r2,3)



def test_write_values_tracks_from_list():

    test_data_dir = shephard.get_data('test_data')
    fasta_file = '%s/%s' % (test_data_dir, 'testset_1.fasta')
    
    P = uniprot.uniprot_fasta_to_proteome(fasta_file)

    # build phosphostring

    r1 = random.randint(1,20)
    r2 = random.random()
    for protein in P:

        new_vals = []
        for i in protein.sequence:

            if i in ['S','T','Y']:
                new_vals.append(r1)
            else:
                new_vals.append(r2)

        protein.add_track('test_track', values=new_vals)

    track_list = []
    for protein in P:
        track_list.append(protein.track('test_track'))

    # write all the tracks out
    interfaces.si_tracks.write_tracks_from_list(track_list, 'output_test/all_values_test_from_list.tsv')

    # read the tracks into a new proteome
    P2 = uniprot.uniprot_fasta_to_proteome(fasta_file)
    interfaces.si_tracks.add_tracks_from_file(P2, 'output_test/all_values_test_from_list.tsv', mode='values')

    # check everything worked ok
    for protein in P2:
        print(protein)
        for i in range(1,len(protein)+1):

            if protein.residue(i) in ['S','T','Y']:
                assert protein.track('test_track').value(i) == r1
            else:
                assert protein.track('test_track').value(i) == np.round(r2,3)



def test_write_symbols_tracks_from_list():

    test_data_dir = shephard.get_data('test_data')
    fasta_file = '%s/%s' % (test_data_dir, 'testset_1.fasta')
    
    P = uniprot.uniprot_fasta_to_proteome(fasta_file)

    # build phosphostring

    r1 = random.randint(1,20)
    r2 = random.random()
    for protein in P:

        new_symbols =''
        for i in protein.sequence:

            if i in ['S','T','Y']:
                new_symbols = new_symbols + "P"
            else:
                new_symbols = new_symbols + "-"

        protein.add_track('test_track', symbols=new_symbols)

    track_list = []
    for protein in P:
        track_list.append(protein.track('test_track'))

    # write all the tracks out
    interfaces.si_tracks.write_tracks_from_list(track_list, 'output_test/all_values_test_from_list.tsv')

    # read the tracks into a new proteome
    P2 = uniprot.uniprot_fasta_to_proteome(fasta_file)
    interfaces.si_tracks.add_tracks_from_file(P2, 'output_test/all_values_test_from_list.tsv', mode='symbols')

    # check everything worked ok
    for protein in P2:
        print(protein)
        for i in range(1,len(protein)+1):

            if protein.residue(i) in ['S','T','Y']:
                assert protein.track('test_track').symbol(i) == 'P'
            else:
                assert protein.track('test_track').symbol(i) == '-'



def test_write_symbols_tracks_from_list():

    test_data_dir = shephard.get_data('test_data')
    fasta_file = '%s/%s' % (test_data_dir, 'testset_1.fasta')
    
    P = uniprot.uniprot_fasta_to_proteome(fasta_file)

    # build phosphostring

    r1 = random.randint(1,20)
    r2 = random.random()
    for protein in P:



        # add symbol
        if random.random() < 0.5:            
            new_symbols =''
            for i in protein.sequence:

                if i in ['S','T','Y']:
                    new_symbols = new_symbols + "P"
                else:
                    new_symbols = new_symbols + "-"

            protein.add_track('test_track_1', symbols=new_symbols)

        # add value
        else:
            new_vals = []
            for i in protein.sequence:

                if i in ['S','T','Y']:
                    new_vals.append(r1)
                else:
                    new_vals.append(r2)

            protein.add_track('test_track_2', values=new_vals)
    
    track_list = []
    for protein in P:
        track_list.append(protein.track('test_track_1', safe=False))
        track_list.append(protein.track('test_track_2', safe=False))

    with pytest.raises(InterfaceException):
        interfaces.si_tracks.write_tracks_from_list(track_list, 'output_test/should_never_be_written.tsv')


def test_write_tracks_sanity_checks():

    with pytest.raises(InterfaceException):
        interfaces.si_tracks.write_tracks_from_list([1,None,2], 'output_test/should_never_be_written.tsv')

    



"""
SHEPHARD: 
Sequence-based Hierarchical and Extendable Platform for High-throughput Analysis of Region of Disorder

Authors: Garrett M. Ginell & Alex S. Holehouse
Contact: (g.ginell@wustl.edu)

Holehouse Lab - Washington University in St. Louis
"""

def get_site_density_in_domain_normalized_by_protein(domain, site_types, sample_size=20, max_enrichment=100, min_enrichment=0.01):
    """

    """
    
    # first calculate the density of sites in the domain
    density_of_sites_in_domain = len(domain.get_sites_by_type(site_types))/len(domain)
    
    subsample=[]

    # next, for sample_size iteractions we're going to...
    count=0
    for i in range(0, sample_size):

        # next we're going to randomly select another region of the protein that
        # is the same size as the domain...
        maxlen = len(domain.protein)
        domain_size = len(domain)    
        d_start = random.randint(0,maxlen-domain_size)        
        d_end   = d_start + domain_size

        # and then count how many hits in the site
        count = count + len(domain.protein.get_sites_by_type_and_range(site_types, d_start, d_end))


    # we do this to avoid dividing by zero - i.e. compute on average how many sites one finds
    # and then use that number to compute the density across the domain)
    mean_expected = (count/sample_size)/len(domain)

    # if we didn't find sites anywhere
    if density_of_sites_in_domain == 0:
        if count == 0:
            return 1
        else:
            return max_enrichment
    if count == 0:
        return min_enrichment

    return density_of_sites_in_domain/mean_expected


def get_site_density_percentile_normalized_by_protein(domain, site_types):

    domain_length = len(domain)

    # get density of sites in the domain
    density_of_sites_in_domain = len(domain.get_sites_by_type(site_types))/domain_length

    # 
    density_vector  = site_tools.build_site_density_vector(domain.protein, site_types=site_types, window_size=domain_length, append_leading_lagging=False)
    density_vector.sort()

    c=0
    for i in density_vector:
        c=c+1
        if i == density_of_sites_in_domain:
            return c/len(density_vector)




    
                
            



"""
SHEPHARD: 
Sequence-based Hierarchical and Extendable Platform for High-throughput Analysis of Region of Disorder

Authors: Garrett M. Ginell & Alex S. Holehouse
Contact: (g.ginell@wustl.edu)

Holehouse Lab - Washington University in St. Louis
"""

from . import general_utilities
from .exceptions import ProteomeException
from .protein import Protein
from itertools import islice
import copy

class Proteome:
    """
    The Proteome object is the main unit for information storage in 
    SHEPHARD.

    There are a few ways that new Proteomes can be generated:

    * By reading in a FASTA file (using shephard.interfaces.apis.fasta)

    The Proteome constructor takes a single argument, which is a 
    list of protein dictionaries or a list of Protein objects. This 
    means Proteome objects can be generated directly (see below for 
    a definition of protein dictionaries). However, it is often more 
    convenient to build Proteomes from FASTA files. For more information 
    on this see the ``api`` documentation.
    
        
    Protein dictionaries are dictionaries that must contain four elements 
    (others are ignored).
    
    * ``sequence`` : *str* - Amino acid sequence of the protein.        

    * ``name`` : *str* - Name of the protein (this can be anything, it is not used internally so no constraints on what this is.

    * ``unique_ID`` :  *str* - This must be unique with respect to all other unique_IDs in the set of proteins in the input list.
                                          
    * ``attributes`` : *dict* - Dictionary of one or more attributes to apply to this protein. Key/value pairs in this dictionary can be arbitrary and are user defined.
    
    

    As an example:

    >>> protein_dictionary_example = {'sequence':'ALAPSLLPAMPALSPALSP', 
                                      'name': 'my protein fragment', 
                                      'unique_ID':'UXX01', 'attributes':{}}
    >>> dictionary_list = []
    >>> dictionary_list.append(protein_dictionary_example)
    >>> P = Proteome(dictionary_list)
                                          
    Note that ``sequence``, ``name`` and ``unique_ID`` are cast to *str* by 
    the function, so if numerical values are passed for any these will be 
    converted to strings.
    
    **Notes**

    * NOTE that ALL FOUR of these are required for EACH protein, even if the 
      attributes dictionary is empty.

    * The unique_ID is checked for uniqueness against all others in the Proteomes 
      and will throw and exception if it is, in fact, not unique.

    * Additional proteins can be added using the `.add_protein()` or 
      `.add_proteins() function.

    """

    ## ------------------------------------------------------------------------
    ##
    def __init__(self, input_list = None, attributes = None, force_overwrite=False):
        # See the Proteome class documentation for constructor info
        """
        Constructor that generates a new Proteome object. This includes 
        taking a list of Protein objects or protein dictionaries (input_list) 
        and, optionally an attributes dictionary for the proteome itself.
                
        In addition, force_overwrite can be used to deal with duplicate 
        entries in the input_list.

        There are two types of lists that can be tolerated when passed to
        the Proteome constructor: a list of protein dictionaries and a 
        list of Protein objects. Note this simply is using the 
        .add_proteins() function to add the passed input_list.

        **Protein dictionaries**        
        One mode of adding multiple proteins is by passing a list of 
        Protein dictionaries.

        Protein dictionaries are dictionaries that posses the following 
        key-value pairs ::

            'sequence'   : amino acid sequence (str)
            'name'       : protein name (str)
            'unique_ID'  : The unique identification number used for the 
                           protein (str)
            'attributes' : A dictionary of arbitrary key-value pairs to 
                           associate with the protein (dict or None)
        
        Additional keys/value pairs are ignored and ALL four of these must
        be included. If any are missing for any protein entry this function 
        raises a ProteomeException.

        **Protein objects**
        A second mode of adding multiple proteins is by passing a list of
        Protein objects. This is useful if you're creating a new Proteome
        based on a subset of proteins from an existing Proteome.

        In both cases, the function automatically determines the type of 
        the passed list, and adds dictionaries accordingly. Note that 
        in both cases proteins are added by value - i.e. a new Protein
        object is generated.
        
        Parameters
        -----------
        input_list : list (default = None)
            List of Protein dictionaries

        attributes : dict (default = None)
            A an arbitrary set of key-value pairs to annotate the proteome 
            with metadata.
        
        force_overwrite : bool (default = False)
            If set to False and there are duplicate unique_IDs in protein 
            dictionaries in the input_list this will trigger an exception. 
            However, if set to True then the 'last' entry overwrites a             
            more recent one in the case of duplicates.

        """

        # initiallize book-keeping instruments
        self._records = {}
        self._unique_domain_types = {}
        self._unique_site_types = {}         # dict that mapes a Site type to a count
        self._unique_track_names = {}        # dict that mapes Track name to count
        self._track_name_to_track_type = {}  # dict that maps Track name to track type ('values' or 'symbols')
        
        # check attributs dictionary
        general_utilities.variable_is_dictionary(attributes, ProteomeException, 'attributes argument passed to proteome is not a dictionary', or_none = True)

        if attributes is None:            
            self._attributes = {}
        else:
            self._attributes = attributes

        # if no/empty input provided then we're done
        if input_list is None or len(input_list) == 0:
            return

        # else try and add proteins - probably could be more soph
        self.add_proteins(input_list)



    ## ------------------------------------------------------------------------
    ##                
    def check_unique_ID(self, unique_id):
        """
        Function that checks if a given unique ID is found. Note that this 
        function is not needed for testing if a unique_ID is present if the 
        goal is to request Protein Objects (or not). Instead, one can use 
        the .protein(<unique_ID>, safe=False). By setting safe=False if the
        unique_ID is not found then this function will simply return None.

        Parameters
        -----------
        unique_id : string
            String corresponding to a unique_ID associated with some protein

        Returns
        ----------
        bool
            Returns True if the passed ID is present, or False if not.

        """
        if unique_id in self._records.keys():
            return True
        else:
            return False



    ## ------------------------------------------------------------------------
    ##
    @property
    def proteins(self):
        """
        Returns a list of unique_IDs that correspond to the proteins in 
        this Proteome. NOTE this returns a list of the IDs, not the actual 
        Protein objects. To get the corresponding protein object one must 
        use the ``.protein(<unique_ID>)`` notation.
                
        Returns
        --------

        ``list`` of ``str``
            Returns a list of unique IDs

        """

        return list(self._records.keys())



    ## ------------------------------------------------------------------------
    ##
    def protein(self, unique_ID, safe=True):
        """
        Returns the ``Protein`` object associated with the passed unique_ID. 
        If there is no Protein associated with the provided unique_ID then 
        if ``safe=True`` (default) an exception is raised, while if 
        ``safe=False`` then ``None`` is returned.

        Parameters
        -----------
        unique_id : string
            String corresponding to a unique_ID associated with some protein

        safe : bool (default = True)
            If set to True then a missing unique_ID  will raise an exception. 
            If ``False`` then a missing unique_ID will simply return None
            
        Returns
        --------
        
        Protein Object, None
            Depending on if the passed unique_ID is found in the ``Proteome``, 
            a ``Protein`` object or None will be returned.            

        """

        # convert unique IDs to strings because this typing also happens 
        # when new proteins are added
        unique_ID_str = str(unique_ID)

        try:
            return self._records[unique_ID_str]
        except KeyError:
            if safe:
                raise ProteomeException("unique_ID '%s' not found in proteome" % (unique_ID))
            else:
                return None



    ## ------------------------------------------------------------------------
    ##
    def add_protein(self, sequence, name, unique_ID, attributes=None, force_overwrite=False):
        """
        Function that allows the user to add a new protein to a Proteomes 
        in an ad-hoc fashion. In general most of the time it will make 
        sense to add proteins all at once from some input source, but the 
        ability to add proteins one at a time is also useful. 

        If a duplicate unique_ID is passed an exception (ProteomeException) 
        is raised.
        
        Parameters
        -----------
        sequence : string
            Amino acid sequence of the protein. Note  - no sanity check of 
            the sequence is performed.

        name : string
            String reflecting the protein name. Again this can be 
            anything.
        
        unique_id : string
            String corresponding to a unique_ID associated with some 
            protein.

        attributes : dict (default = None)
            The attributes dictionary provides a key-value pairing for 
            arbitrary information. This could include gene names, different 
            types of identifies, protein copy number, a set of protein 
            partners, or anything else one might wish to associated with the
            protein as a whole. Default is None.

        force_overwrite : Bool (default = False)
            If set to False and a unique_ID is included that already is 
            found then this function will raise an exception. However, if 
            set to True it will automatically overwrite the pre-existing 
            entry. (Default = False).
           
        Returns
        --------        
        None
            No return status, but valid proteins included in the input_list 
            will be added to to the underlying proteome.

        """
        
        unique_ID_str = str(unique_ID)

        if unique_ID_str in self._records:
            if force_overwrite is False:
                raise ProteomeException('Non-unique unitque_ID passed [%s]' % (unique_ID_str))

        self._records[unique_ID_str] = Protein(sequence, name, self, unique_ID_str, attributes)



    ## ------------------------------------------------------------------------
    ##        
    def add_proteins(self, input_list, force_overwrite=False):
        r"""
        Function that allows the user to add a multiple new proteins using 
        either a list of protein dictionaries (described below) or a list 
        of Protein objects.        

        **Protein dictionaries**
        
        One mode of adding multiple proteins is by passing a list of 
        Protein dictionaries.

        Protein dictionaries are dictionaries that posses the following 
        key-value pairs ::

            'sequence'   : amino acid sequence (str)
            'name'       : protein name (str)
            'unique_ID'  : The unique identification number used for the 
                           protein (str)
            'attributes' : A dictionary of arbitrary key-value pairs to 
                           associate with the protein (dict or None)
        
        Additional keys/value pairs are ignored and ALL four of these must
        be included. If any are missing for any protein entry this function 
        raises a ProteomeException.

        **Protein objects**
        A second mode of adding multiple proteins is by passing a list of
        Protein objects

        In both cases, the function automatically determines the type of 
        the passed list, and adds dictionaries accordingly. Note that 
        in both cases proteins are added by value - i.e. a new Protein
        object is generated.
        
        Parameters
        -----------
        input_list : list
            List of Protein dictionaries or list of Protein objects
        
        force_overwrite : bool (default = False)
            If set to False and a unique_ID is included that already is 
            found then this function will raise an exception. However, if 
            set to True it will automatically overwrite the pre-existing 
            entry. 

        Returns
        --------
        
        None
            No return status, but valid proteins included in the input_list 
            will be added to to the underlying proteome.
            

        """

        # cycles over every element in the input list and builds a new
        # list where each 
        type_list = list(set([type(i) for i in input_list]))

        # checks if only one type of object is found here
        if len(type_list) > 1:
            raise ProteomeException(f'Trying to add Proteins to a Proteome and the input_list contains more than one type {type_list}')

        # if we're using a list of protein dictionaries
        if type_list[0] == dict:
            self._add_proteins_dict(input_list, force_overwrite)

        # if we're using a list of Protein objects
        elif type_list[0] == Protein:
            self._add_proteins_Protein(input_list, force_overwrite)

        

    ## ------------------------------------------------------------------------
    ##        
    def _add_proteins_dict(self, input_list, force_overwrite=False):
        """
        Internal function that mirrors add_proteins() but operates if every
        element in the input_list is a dictionary.

        Importantly, this works by create a NEW proteins and populating
        based on the key-value mapping in the protein dictionary. 

        Protein dictionaries are dictionaries that posses the following 
        key-value pairs:

            'sequence'   : amino acid sequence (str)
            'name'       : protein name (str)
            'unique_ID'  : The unique identification number used for the 
                           protein (str)
            'attributes' : A dictionary of arbitrary key-value pairs to 
                           associate with the protein (dict or None)


        This function is not be called directly, but instead used by 
        add_proteins() function

        Parameters
        --------------
        input_list : list 
            List of dictionaries  

        force_overwrite : bool, default=False
            Flag which if set to True will mean a Protein in the input_list
            would overwrite an existing protein with the same unique_ID.

        Returns
        --------
        None
            No return type, but will add the protein dictoinaries in the 
            input_list into this Proteome.

        
        """
        
        # for each protein entry in the input list
        for entry in input_list:
        
            try:
                sequence       = str(entry['sequence'])
                name           = str(entry['name'])
                unique_ID      = str(entry['unique_ID'])
                attributes     = entry['attributes']
            except KeyError:
                # if something goes wrong while extracting the four required attributes we build a 
                # diagnosis string and then print this as we raise an exception. The goal here is to
                # try and provide the user with as much info as possible to diagnose the problem
            
                diagnosis_string = self.__build_diagnosis_string_proteome_construction(entry)
                raise ProteomeException('%s'%(diagnosis_string))
            
            if unique_ID in self._records:
                if force_overwrite is False:
                    raise ProteomeException('Non-unique unique_ID passed [%s]' % (unique_ID))

            # add in a new protein
            self._records[unique_ID] = Protein(sequence, name, self, unique_ID, attributes)


    ## ------------------------------------------------------------------------
    ##        
    def _add_proteins_Protein(self, input_list, force_overwrite=False):
        """
        Internal function that mirrors add_proteins() but operates if every
        element in the input_list is a Protein object.

        Importantly, this works by create a NEW protein, and ensure all 
        complex datatypes associated with that new protein are copied. 

        This function is not be called directly, but instead used by 
        add_proteins() function

        Parameters
        --------------
        input_list : list of shephard.protein.Protein objects
            List of Protein objects to be copied into this Proteome

        force_overwrite : bool, default=False
            Flag which if set to True will mean a Protein in the input_list
            would overwrite an existing protein with the same unique_ID.

        Returns
        --------
        None
            No return type, but will add the Protein objects in the input_list
            into this Proteome.
        
        """
        
        # for each protein entry in the input list
        for entry in input_list:
            
            # get the unique ID and, if this ID is already found in the Proteome
            # raise an exception UNLESS force_overwrite is True
            unique_ID = entry.unique_ID
            if unique_ID in self._records:
                if force_overwrite is False:
                    raise ProteomeException('Non-unique unique_ID passed [%s]' % (unique_ID))


            ##
            ## New protein is fully created and all complex data types are copied, so this
            ## new protein is a completely distinct entity to the protein in the original
            ## input list. This avoids any possible issues with cross-referencing back against
            ## old data structures and keeps things clean. Also ensures that the new proteome
            ## has updated unique sites and unique domains lists in the usual way. Basically
            ## this is the most appropritae way to add an existing protein to a new proteome
            ##
            # create new protein
            new_protein = Protein(entry.sequence, entry.name, self, entry.unique_ID)

            # update attributes
            for a in entry.attributes:
                new_protein.add_attribute(a, copy.deepcopy(entry.attribute(a)))

            # update domains
            for d in entry.domains:
                new_protein.add_domain(d.start, d.end, d.domain_type, copy.deepcopy(d._attributes))

            # update sites
            for s in entry.sites:
                new_protein.add_site(s.position. s.site_type, s.symbol, s.value, copy.deepcopy(s._attributes))

            # update tracks
            for t in entry.tracks:
                vals = t.values

                # if vals present then we're creating a values track. Note we can get away with
                # a shallow copy because we know these tracks will only have ints or chars in their
                # elements, which are appropriately copied by a shallow copy
                if vals:                    
                    new_protein.add_track(t.name, vals.copy(), None)
                else:
                    new_protein.add_track(t.name, None, t.symbols.copy())

            self._records[unique_ID] = new_protein
           
            


    ## ------------------------------------------------------------------------
    ##        
    def remove_protein(self, unique_ID, safe=True):
        """
        Function that removes a given protein from the Proteome based on the
        passed unique_ID. If the passed unique_ID does not exist then this 
        will trigger an exception unless safe=False.

        
        Parameters
        ------------
        unique_ID : str
            Unique ID that will be used to retrieve a given protein

        safe : bool (default = True)
            Flag that if set to True means if a passed unique_ID is missing 
            from the underlying proteome object an exception wll be raised 
            (ProteomeException). If set to False, a missing unique_ID is 
            ignored.


        Returns
        -----------
        None
            No return type but will remove an entry from the Proteome.
           
        """

        unique_ID_str = str(unique_ID)

        if unique_ID_str in self._records:
            del self._records[unique_ID_str]
        else:
            if safe:
                raise ProteomeException('Passed unique_ID [%s] not found in this proteome' % (unique_ID_str))
            


    ## ------------------------------------------------------------------------
    ##        
    def remove_proteins(self, input_list, safe=True):
        """
        Function that removes a given proteome from the Proteome based on
        the passed unique_ID. If the passed unique_ID does not exist then 
        this will trigger an exception unless safe = False.

        Parameters
        ------------
        input_list : list of str
            List that contains the unique IDs that will be used to select 
            proteins for deletion.

        safe : bool (default = True)
            Flag that if set to True means if a passed unique_ID is missing 
            from the underlying proteome object an exception wll be raised 
            (ProteomeException). If False a missing unique_ID is ignored.

        Returns
        -----------
        None
            No return type but will remove an entry from the proteome           

        """

        for unique_ID in input_list:
            self.remove_protein(unique_ID, safe=safe)
            


     
    ###################################
    ##                               ##
    ##     ATTRIBUTE FUNCTIONS       ##
    ##                               ##
    ###################################


    ## ------------------------------------------------------------------------
    ##
    @property
    def attributes(self):
        """
        Provides a list of the keys associated with every attribute associated
        with this protein.

        Returns
        -------
        list
            returns a list of the attribute keys associated with the Proteome. 


        """
        return list(self._attributes.keys())
    
    
    ## ------------------------------------------------------------------------
    ##
    def attribute(self, name, safe=True):

        """
        Function that returns a specific attribute as defined by the name. 

        Recall that attributes are name : value pairs, where the 'value' 
        can be anything and is user defined. This function will return 
        the value associated with a given name.

        Parameters
        ----------------
        name : str
             The attribute name. A list of valid names can be found by 
             calling the ``<Proteome>.attributes()`` (which returns a list 
             of the valid names).
             
        safe : bool (default = True)
            Flag which if true with throw an exception if an attribute with 
            the same name already exists.
            
        Returns
        ---------
        Unknown 
            Will either return whatever was associated with that attribute 
            (which could be anything) or None if that attribute is missing.            
        
        """

        # if name is in the _atributes dictionary the  return
        if name in self._attributes:
            return self._attributes[name]
        else:

            # else if safe was passed raise an exception if that attribute was missing
            if safe:
                raise ProteomeException('Requesting attribute [%s] from Proteome [%s] but this attribute has not been assigned' % (name, str(self))) 

            # if safe not passed just return None
            else:
                return None
                


    ## ------------------------------------------------------------------------
    ##
    def add_attribute(self, name, val, safe=True):
        """
        Function that adds an attribute. Note that if safe is true, 
        this function will raise an exception if the attribute is 
        already present. If safe=False, then an exisiting value will 
        be overwritten.

        Parameters
        ----------------

        name : str
            The parameter name that will be used to identify it

        val : <anything>
            An object or primitive we wish to associate with this 
            attribute

        safe : bool (default = True)
            Flag which if True with throw an exception if an attribute 
            with the same name already exists, otherwise the newly 
            introduced attribute will overwrite the previous one.

        Returns
        ---------
            None - but adds an attribute to the calling object

        """

        if safe:
            if name in self._attributes:
                raise ProteomeException("Trying to add attribute [%s=%s] to Proteome [%s] but this attribute is already set.\nPossible options are: %s" %(name,val, str(self), str(self._attributes.keys())))
                
        self._attributes[name] = val        
        

    ## ------------------------------------------------------------------------
    ##
    def remove_attribute(self, name, safe=True):
        """
        Function that removes a given attribute from the Proteome based on the 
        passed attribute name. If the passed attribute does not exist or is not 
        associate with the protein then this will trigger an exception 
        unless safe=False.

        Parameters
        ----------------

        name : str
            The parameter name that will be used to identify it

        safe : bool (default = True)
            Flag which if True with throw an exception if an 
            attribute this name does not exists. If set to
            False then if an attribute is not found it is simply
            ignored
            
        Returns
        ---------
        None
            No return type but will remove an attribute from the 
            Proteome if present.
            
        """

        if name not in self._attributes:
            if safe:
                raise ProteomeException(f'Passed attribute [{name}] not found in {self}')
        else:
            del self._attributes[name]



    ###################################
    ##                               ##
    ##       DOMAIN FUNCTIONS        ##
    ##                               ##
    ###################################

    ## ------------------------------------------------------------------------
    ##
    @property
    def domains(self):
        """
        Function that returns a list of all domain objects associated with
        the Proteome. 

        This function is useful if you wish to indiscriminately ask questions 
        of domains without considering the proteins they come from. However, 
        each Domain has a Protein object associated with it (via the .protein 
        operator), so one can always map a Domain back to a Protein.

        Returns
        --------------
        list of Domains
             A list of all the Domains from every protein in the Proteome
        
        """

        all_domains = []
        for prot in self:
            all_domains.extend(prot.domains)

        return all_domains


        
    ## ------------------------------------------------------------------------
    ##                
    @property
    def unique_domain_types(self):
        """
        Returns the list of unique Domain types associated with this Proteome.

        Return
        -------

        list of str
            Each element in the list is a string that corresponds to a Domain 
            type.

        """

        # Some description of what's going on here is in order. Every time a new domain
        # is added, the Domain constructor calls the function _Domain__update_domain_types
        # which checks if the domain_type of the domain being added is already in the
        # _unique_domain_types list. If yes, fine, if no, it gets added. This means
        # _unique_domain_types keeps track of a count of the complete number of unique
        # domains in the Proteome. An analogous setup holds true for the sites and tracks.
        
        return list(self._unique_domain_types.keys())


    ## ------------------------------------------------------------------------
    ##
    def get_domains_by_type(self, domain_type, perfect_match=True):
        """
        Function that returns a list of domains from all proteins that matched against
        a specific domain type name. 
        
        Parameters
        ------------
        domain_type : string
            String associated domain_type that you want to search for.

        perfect_match : bool (default = True)
            Flag that identifies if the domain names should be a perfect 
            match (=True) or if the string passed should just appear 
            somewhere in the domain_type string
            
        Returns
        -----------
        list
            Returns a list of Domain objects that match the requested type. 
            Objects are ordered by starting position in sequence.
                                
        """
        return_list = []
        for p in self:
            return_list.extend(p.get_domains_by_type(domain_type, perfect_match))

        return return_list



    ###################################
    ##                               ##
    ##        SITES FUNCTIONS        ##
    ##                               ##
    ###################################

    ## ------------------------------------------------------------------------
    ##                
    @property
    def sites(self):
        """
        Function that returns a list of all Site objects associated with 
        the Proteome. 

        This function is useful if you wish to indiscriminately ask questions 
        of sites without considering the proteins they come from. However, 
        each Site has a Protein object associated with it (via .protein 
        operator), so one can always map a Site back to a Protein.

        Returns
        --------------
        list of Sites
             A list of all the Sites from every protein in the Proteome
        
        """

        all_sites = []

        for prot in self:
            all_sites.extend(prot.sites)
        return all_sites
                    

    ## ------------------------------------------------------------------------
    ##                
    @property
    def unique_site_types(self):
        """
        Returns the list of unique Site types associated with this Proteome.

        Return
        -------

        list of str
            Each element in the list is a string that corresponds to a Site type

        """

        # Some description of what's going on here is in order. Every time a new site
        # is added, the Site constructor calls the function _Site__update_site_types
        # which checks if the domain_type of the domain being added is already in the
        # _unique_site_types list. If yes, fine, if no, it gets added. This means
        # _unique_site_types keeps track of a count of the complete number of unique
        # sites in the Proteome. An analogous setup holds true for domains and tracks.

        return list(self._unique_site_types.keys())
    

    ## ------------------------------------------------------------------------
    ##
    def get_sites_by_type(self, site_types):
        """
        Function that returns a list of sites from all proteins that matched against
        a specific site type name or set of site type names.
        
        Parameters
        ------------

        site_types : string or list of strings
            One or more possible site_types that may be found in the protein. 
            Either a single string or a list of strings can be passed, 
            allowing for one or more sites to be grouped together

            
        Returns
        -----------
        list
            Returns a list of Domain objects that match the requested type. 
            Objects are ordered by starting position in sequence.
                                
        """
        return_list = []
        for p in self:
            return_list.extend(p.get_sites_by_type(site_types, return_list=True))

        return return_list



    ###################################
    ##                               ##
    ##        TRACK FUNCTIONS        ##
    ##                               ##
    ###################################

    ## ------------------------------------------------------------------------
    ##                
    @property
    def unique_track_names(self):
        """
        Returns the list of unique Track names associated with this Proteome.

        Return
        -------

        list of strings
            Each element in the list is a string that corresponds to a 
            Track name found in one (or more) proteins

        """

        # Some description of what's going on here is in order. Every time a new track
        # is added, the Track constructor calls the function _Track__update_track_names
        # which checks if the track_name of the domain being added is already in the
        # _unique_track_types list. If yes, fine, if no, it gets added. This means
        # _unique_track_names keeps track of a count of the complete number of unique
        # domains in the Proteome. An analogous setup holds true for sites and domains
        
        
        return list(self._unique_track_names.keys())

    ## ------------------------------------------------------------------------
    ##                
    @property
    def track_names_to_track_type(self):
        """
        Returns a (copy of a) dictionry that maps track name to track type. 
        We return a copy so there's no way we can accidentally break the 
        internal book-keeping of the Proteome object.

        Return
        -------
        dict
            A dictionary that contains the unique track names and maps 
            each name to either values or type.
        
        """

        return dict(self._track_name_to_track_type)


    ####################################
    ##                                ##
    ##       INTERNAL FUNCTIONS       ##
    ##                                ##
    ####################################

    ## ------------------------------------------------------------------------
    ##                
    def __len__(self):
        """
        The length of the Proteome is defined as the number of proteins 
        in it.
        
        Returns
        -------
        int
            Returns an integer that reflects the number of proteins
        """

        # this function means when we call len(Proteome_Object) we get back
        # the number of proteins in it

        return len(self._records)



    ## ------------------------------------------------------------------------
    ##                
    def __repr__(self):
        """
        Provides a nice representation of the Proteome
        
        Returns
        -------
        string
            Formatted description of the Proteome.

        """

        # this function means when we print a Proteome object or cast it to 
        # a string we get a nice/informative representation, rather than the
        # id of the object

        return "[Proteome]: Sequence dataset with %i protein records" %(len(self))



    ## ------------------------------------------------------------------------
    ##                
    def __iter__(self):
        """
        Allows a Proteome object to act as a generator that yields actual 
        proteins, so the syntax
        
        .. code-block:: python

           for protein in ProteomeObject:
               print(protein.sequence)
        
        is be valid and would iterate through the proteins in the Proteome. 
        
        This makes performing some analysis over all proteins quite easy.

        """

        for i in self._records:
            yield self._records[i]

    ## ------------------------------------------------------------------------
    ##                
    def __contains__(self, m):
        """
        Enables the syntax X in Proteome to be used, where X can be 
        either a unique ID or a Proteome object.

        .. code-block:: python

           if protein.unique_ID in ProteomeObject:
               print(f'The protein {protein} is in the Proteome!')


        """

        if type(m) == str:
            if m in self._records.keys():
                return True
            else:
                return False

        elif type(m) == Protein:
            if m.unique_ID in self._records.keys():
                return True
            else:
                return False

    ## ------------------------------------------------------------------------
    ##                        
    def __getitem__(self, key):
        """
        Allows slicing index into Proteome to retrieve subsets of protein

        .. code-block:: python

           first_protein = ProteomeObject[0]
           print(f'The first protein is {first_protein}')

        """


        if isinstance(key, int) and key >= 0:
            return list(islice([self._records[i] for i in self._records], key, key+1))[0]

        elif isinstance(key, slice):
            return list(islice([self._records[i] for i in self._records], key.start, key.stop, key.step))
        else:
            raise KeyError("Key must be non-negative integer or slice, not {}"
                           .format(key))


    ## ------------------------------------------------------------------------
    ##                
    def _Domain__update_domain_types(self, domain_type):
        """
        INTERNAL FUNCTION (not for public API use)

        Note - we this function is named as __Domain_... so it can be 
        specifically and uniquely be called from a Domain object. This 
        function is ONLY called last thing in the Domain constructor 
        where it allows the Proteome object to keep track of the total 
        number of unique domain types in the Proteome.

        The function is (by default) called by the Domain constructor.
        
        Parameters
        ----------------
        domain_type : string
            String that defines a domain type

        Returns
        ---------------

        No return value, but will appropriately update the Proteome object

        """
        # because _unique_track_names is a dictionary this scales O(1) with number 
        # of track names
        if domain_type not in self._unique_domain_types:
            self._unique_domain_types[domain_type] = 1
        else:
            self._unique_domain_types[domain_type] = self._unique_domain_types[domain_type] + 1

    ## ------------------------------------------------------------------------
    ##                
    def _Protein__decrement_domain_types(self, domain_type):
        """
        INTERNAL FUNCTION (not for public API use)
        
        Note - we this function is named as __Protein_... so it 
        can be specifically and uniquely be called from a Protein 
        object. This function is ONLY called last thing when a 
        Protein object deletes a domain

        Parameters
        ----------------
        track_name : string
            String that defines a Domain type

        Returns
        ---------------

        No return value, but will appropriately update the Proteome 
        object

        """

        # if we can't find the domain name in the unique domain types... this is bad!
        if domain_type not in self._unique_domain_types:
            raise ProteomeException("Tried to remove a Domain type [{domain_type}] from the Proteome.unique_domain_types dictionary but the Domain type could not be found. This is a bug. Please report as a GitHub Issue.")
            
        # we are removing a unique domain_type name! Big event!
        elif self._unique_domain_types[domain_type] == 1:
            del self._unique_domain_types[domain_type]
        # else decrement one
        else:
            self._unique_domain_types[domain_type] = self._unique_domain_types[domain_type] - 1


    ## ------------------------------------------------------------------------
    ##                
    def _Track__update_track_names(self, track_name, track_type):
        """
        INTERNAL FUNCTION (not for public API use)
        
        Note - we this function is named as __Track_... so it can be 
        specifically and uniquely be called from a Track object. This 
        function is ONLY called last thing in the Track constructor 
        where it allows the Proteome object to keep track of the total 
        number of unique Track types in the Proteome.

        The function is (by default) called by the Track constructor

        Parameters
        ----------------
        track_name : string
            String that defines the Track name 

        Returns
        ---------------

        No return value, but will appropriately update the Proteome object

        """

        # because _unique_track_names is a dictionary this scales O(1) with number 
        # of track names
        if track_name not in self._unique_track_names:
            self._unique_track_names[track_name] = 1
            self._track_name_to_track_type[track_name] = track_type

        else:
            self._unique_track_names[track_name] = self._unique_track_names[track_name] + 1
            if self._track_name_to_track_type[track_name] != track_type:
                raise ProteomeException(f"Tried to assigned track name [{track_name}] as a [{track_type}] track, but this track was already assigned as a [{self._track_name_to_track_type[track_name]}] track. Cannot have two tracks with the same name but different types")
            

    ## ------------------------------------------------------------------------
    ##                
    def _Protein__decrement_track_names(self, track_name):
        """
        INTERNAL FUNCTION (not for public API use)
        
        Note - we this function is named as __Protein_... so it can be
        specifically and uniquely be called from a Protein object. 
        This function is ONLY called last thing when a Protein object 
        deletes a Track
                

        Parameters
        ----------------
        track_name : string
            String that defines the Track name 

        Returns
        ---------------

        No return value, but will appropriately update the Proteome 
        object.

        """

        # if we can't find the track name in the unique track names... this is bad!
        if track_name not in self._unique_track_names:
            raise ProteomeException("Tried to remove a Track name [{track_name}] from the Proteome.unique_track_names dictionary but the track could not be found. This is a bug. Please report as a GitHub Issue.")
            
        # we are removing a unique track name! Big event!
        elif self._unique_track_names[track_name] == 1:
            del self._unique_track_names[track_name]
            del self._track_name_to_track_type[track_name]

        # else decrement one
        else:
            self._unique_track_names[track_name] = self._unique_track_names[track_name] - 1


    ## ------------------------------------------------------------------------
    ##                
    def _Site__update_site_types(self, site_type):
        """
        INTERNAL FUNCTION (not for public API use).

        Note - we this function is named as __SITE_... so it can be 
        specifically and uniquely be called from a Site object. This 
        function is ONLY called last thing in the Site constructor where
        it allows the Proteome object to keep track of the total number 
        of unique Site types in the Proteome.

        

       
        The function is (by default) called by the Site constructor.

        Parameters
        ----------------
        site_type : string
            String that defines a site type

        Returns
        ---------------

        No return value, but will appropriately update the Proteome object.

        """

        # because _unique_track_names is a dictionary this scales O(1) with number 
        # of track names
        if site_type not in self._unique_site_types:
            self._unique_site_types[site_type] = 1

        else:
            self._unique_site_types[site_type] = self._unique_site_types[site_type] + 1


    ## ------------------------------------------------------------------------
    ##                
    def _Protein__decrement_site_types(self, site_type):
        """
        INTERNAL FUNCTION (not for public API use)

        
        Note - we this function is named as __Protein_... so it can be specifically
        and uniquely be called from a Protein object. This function is ONLY called
        last thing when a Protein object deletes a Site

        Parameters
        ----------------
        track_name : string
            String that defines a site type

        Returns
        ---------------

        No return value, but will appropriately update the Proteome object

        """

        # if we can't find the track name in the unique track names... this is bad!
        if site_type not in self._unique_site_types:
            raise ProteomeException("Tried to remove a Site type [{site_type}] from the Proteome.unique_site_types dictionary but the Site type could not be found. This is a bug. Please report as a GitHub Issue.")
            
        # we are removing a unique site_type name! Big event!
        elif self._unique_site_types[site_type] == 1:
            del self._unique_site_types[site_type]
        # else decrement one
        else:
            self._unique_site_types[site_type] = self._unique_site_types[site_type] - 1



    ## ------------------------------------------------------------------------
    ##                
    def __build_diagnosis_string_proteome_construction(self, entry):
        """
        INTERNAL FUNCTION (not for public API use)
        
        Function that builds a string to help in diagnosing what might go 
        wrong during Proteome construction. Called by the Proteome constructor.
        
        
        """

        ds = "Error building proteome when parsing the following entry:\n"


        # check the sequence
        try:
            s = str(entry['sequence'])
        except Exception:
            s = 'FAILED'

        ds = ds + "sequence: %s\n" %(s)


        # check the name
        try:
            s = str(entry['name'])
        except Exception:
            s = 'FAILED'

        ds = ds +"name: %s\n" %(s)


        # check the unique_ID
        try:
            s = str(entry['unique_ID'])
        except Exception:
            s = 'FAILED'

        ds = ds +"unique_ID: %s\n" %(s)


        # 
        try:
            s = str(entry['attributes'])
        except Exception:
            s = 'FAILED'

        ds = ds +"attributes: %s\n" %(s)

        return ds




"""
SHEPHARD: 
Sequence-based Hierarchical and Extendable Platform for High-throughput Analysis of Region of Disorder

Authors: Garrett M. Ginell & Alex S. Holehouse
Contact: (alex.holehouse@wustl.edu, g.ginell@wustl.edu)

Holehouse Lab - Washington University in St. Louis
"""

from shephard.exceptions import InterfaceException


## ------------------------------------------------------------------------
##
def check_protein(p, function_name):
    """
    Function that takes takes in some object and tests if its a Protein
    object (or not). If yes returns None, but if no raise an exception with 
    an error message that includes the name of the parent function that is 
    calling this validation function

    Parameters
    --------------
    p : Protein object 
        Unknown object that will be tested to see if it's a Proteome object
        or not

    function_name : string
        Passed string that makes it easy for the user to debug which function 
        has failed

    Returns
    ---------
    None
        No returns but the function will raise an exception if the passed object 
        is not a Proteome object
    

    """
    if "<class 'shephard.protein.Protein'>" == str(p.__class__):
        return None
    else:
        raise InterfaceException(f'First argument passed to function {function_name} was not a Protein object')


## ------------------------------------------------------------------------
##
def check_proteome(p, function_name):
    """
    Function that takes takes in some object and tests if its a Proteome
    object (or not). If yes returns None, but if no raise an exception with 
    an error message that includes the name of the parent function that is 
    calling this validation function

    Parameters
    --------------
    p : Proteome object 
        Unknown object that will be tested to see if it's a Proteome object
        or not

    function_name : string
        Passed string that makes it easy for the user to debug which function 
        has failed

    Returns
    ---------
    None
        No returns but the function will raise an exception if the passed object 
        is not a Proteome object
    

    """
    if "<class 'shephard.proteome.Proteome'>" == str(p.__class__):
        return None
    else:
        raise InterfaceException(f'First argument passed to function {function_name} was not a Proteome object')

## ------------------------------------------------------------------------
##
def check_domain(d, function_name):
    """
    Function that takes takes in some object and tests if its a Domain
    object (or not). If yes returns None, but if no raise an exception with 
    an error message that includes the name of the parent function that is 
    calling this validation function

    Parameters
    --------------
    d : Domain object 
        Unknown object that will be tested to see if it's a Domain object
        or not

    function_name : string
        Passed string that makes it easy for the user to debug which function 
        has failed

    Returns
    ---------
    None
        No returns but the function will raise an exception if the passed object 
        is not a Domain object
    

    """
    if "<class 'shephard.domain.Domain'>" == str(d.__class__):
        return None
    else:
        raise InterfaceException(f'First argument passed to function {function_name} was not a Domain object')


## ------------------------------------------------------------------------
##
def check_track(t, function_name):
    """
    Function that takes takes in some object and tests if its a Track
    object (or not). If yes returns None, but if no raise an exception with 
    an error message that includes the name of the parent function that is 
    calling this validation function

    Parameters
    --------------
    t : Track object 
        Unknown object that will be tested to see if it's a Track object
        or not

    function_name : string
        Passed string that makes it easy for the user to debug which function 
        has failed

    Returns
    ---------
    None
        No returns but the function will raise an exception if the passed object 
        is not a Track object
    

    """
    if "<class 'shephard.track.Track'>" == str(t.__class__):
        return None
    else:
        raise InterfaceException(f'First argument passed to function {function_name} was not a track object')


## ------------------------------------------------------------------------
##
def check_site(s, function_name):
    """
    Function that takes takes in some object and tests if its a Site
    object (or not). If yes returns None, but if no raise an exception with 
    an error message that includes the name of the parent function that is 
    calling this validation function

    Parameters
    --------------
    s : Site object 
        Unknown object that will be tested to see if it's a Site object
        or not

    function_name : string
        Passed string that makes it easy for the user to debug which function 
        has failed

    Returns
    ---------
    None
        No returns but the function will raise an exception if the passed object 
        is not a Site object
    

    """
    if "<class 'shephard.site.Site'>" == str(s.__class__):
        return None
    else:
        raise InterfaceException(f'First argument passed to function {function_name} was not a site object')
    


## ------------------------------------------------------------------------
##
def full_clean_string(instring):
    r"""
    Wrapper function that takes in a string (or a variable that can be
    cast to a string) and ensure it contains neither tab nor colon characters,
    both of which would render a attribute or free-form text string in a 
    SHEPHARD file invalid.

    This function actually calls `clean_string()` twice with parameters to
    replace, ':' and '\\t' characters.

    Parameters
    -----------
    instring : {str, str-castable}
        String (or string castable) variable to be checked

    Returns
    ---------
    str
        Returns a string which will be essentially identical to the input 
        string but cleaned up to remove tab and colon characteracs .

    """

    s = clean_string(instring)
    s = clean_string(s, ':', '-')
    return s



## ------------------------------------------------------------------------
##
def clean_string(instring, delimiter='\t', replace_char=' '):
    r"""
    Function that takes in a variable  that is to be written to a SHEPHARD 
    complient file and ensures it has no delimiter characters in it. This 
    avoids the scenario where - for example - your protein name has a 
    tab in it which introduces a new field into the SHEPHARD data file.

    In general, we assume instring will be a string (str). However, if it
    is not this function also casts it to a string, so that does not need
    to be done prior to a variable being passed in.

    Furthermore, we'd generally recommend any string that will be written
    out as a user-inputable string to a SHEPHARD file be passed through
    this

    If such delimiter characters are found they're replaced by a replace_char, 
    which by default is just a space    
    
    Parameters
    ---------------
    instring : str
        String (or string castable) variable to be checked

    delimiter : str, default='\\t'
        Character or string that we do NOT want to find in the string

    replace_char : str, default=' '
        Character or string that, if a delimiter is found, will 
        replace the  delimiter

    Returns
    ---------
    str
        Returns a string which will be essentially identical to the input 
        string but cleaned up to remove any bad delimiters if they exist.
        

    """

    if type(instring) is not str:
        instring = str(instring)

    return instring.replace(delimiter, replace_char)



## ------------------------------------------------------------------------
##
def parse_key_value_pairs(split_line, filename, linecount, line):
    """
    Helper function for parsing input files that have attributes (for 
    Domains and Sites).

    The function takes in a list of key-value pairs (where key-values 
    are split by a ':' symbol) and returns a parsed dictionary. Note 
    that values will always be strings.
    
    Parameters
    -------------
    split_line : list of strings
        Each string in the list should have the format <KEY> : <VALUE> 

    filename : string
        Name of the file the calling function is parsing. Only used when 
        raising an exception.

    linecount : int
        Current line number the file processing is on. Only used when 
        raising an exception.

    line : string
        Full line that the file processing is on. Again, only used when 
        raising an exception

    """

    attributes = {}

    for idx in range(len(split_line)):
        try:
            sentry = split_line[idx].split(':')
        except Exception:
                
            # should update this to also display the actual error...
            raise InterfaceException(f'Failed parsing key-value pairs in file [{filename}] on line [{linecount}]... line printed below:\n{line}.\nPassed split_line:[{split_line}]')
            
        # note we enclose this in a try/catch block
        try:
            k = sentry[0].strip()
            v = sentry[1].strip()
                        
            attributes[k] = v
        except IndexError:
            raise InterfaceException(f'Failed parsing key-value pairs in file [{filename}] on line [{linecount}]... line printed below:\n{line}.\nPassed split_line:[{split_line}]')


    return attributes


## ------------------------------------------------------------------------
##
def is_comment_line(line):
    """
    Function that checks if a line should be skipped because it's a comment 
    line or not.

    Parameters
    -------------
    line : str
        A line from an input file

    Returns
    -----------
    bool
        If the line starts with a '#' character then the function returns
        True, else returns False


    """
    line = line.strip()

    if line[0] == '#':
        return True
        

            
